# UTILITY TO SHOW DISK USAGE BY BLUE IRIS CLIPS. ALERTS, AND OTHER FILES
# ... BY CAMERA AND FILE TYPE (BVR, MP4, JPG, DAT)
#
# Script created by jaydeel at ipcamtalk.com
#
# BEFORE USING THIS SCRIPT:
#   1. EDIT ALL VARIABLES IN 'USER SETTINGS' SECTION BELOW
#   2. COMMENT OUT UNUSED/UNNEEDED LINES
#
# Version: 8
# Created by jaydeel at ipcamtalk.com
#
# Updates:
# V8   - 2025-02-01 - NEW: added stats for the quantity of READ-ONLY (protected) files in each folder
#                     HEADS-UP: if you routinely protect clips/alerts, then you should pay attention to the  
#                     ratio of READ-ONLY files and increase the folder storage allocation accordingly
#                   - NEW: added timezone designator to the oldest and newest date columns
#                     HEADS-UP: Windows File Explorer displays file timestamps in the CURRENT local timezone; 
#                     therefore the time shown for old files may be off by an hour from their actual created time
#                   - NEW: added argument -NoTimestamp to override the time stamping of the output CSV files
#                   - NEW: added argument -NoPruning to disable the pruning of old CSV files
#                   - NEW: added argument -ClearArchive to delete archived (old) CSV files on demand
#                   - NEW: script now checks if the commandline contains unsupported argument(s) 
#                   - CHANGED: now showing file CreationTime (previously LastWriteTime) in output table's 'Oldest' & 'Newest' date columns
#                   - CHANGED: now showing time as HH:MM instead of HH:MM:SS in output table's 'Oldest' & 'Newest' date columns 
#                   - CHANGED: eliminated redundant TOTAL row from output tables if the folder contains only one item
#                   - CHANGED: eliminated 'Oldest' & 'Newest' entries from the output table's TOTAL row
#                   - CHANGED: function PurgeOldFiles() now gets existing files ordered oldest to newest
#                   - FIXED: bug preventing the TOTAL row from always being displayed at the top of each table
#                   - FIXED: bug for sub-folders containing only a single file
#                   - FIXED: bug in function RegGetVal() (zero-valued key names)
# V7.1 - 2024-05-17 - BREAKING CHANGE: additional user-settings; change to a user setting
#                   - NEW: added user setting '$experienced_user_flg' to hide output intended for first-time users
#                   - CHANGED: user setting '$archive_lifetime_days' now supports negative integer (prunes all archive files)
# V7   - 2024-05-16 - BREAKING CHANGE: additional user-settings 
#                   - NEW: added argument -ShowFolders to display all Blue Iris Storage folders & highlight those that will be displayed
#                   - NEW: script now self-checks if ANY camera is using a non-default filename format (Camera settings > Record tab);
#                     Note: this feature is skipped if user setting 'default_filename_format_flg' is detected (regardless of its value)
#                   - NEW: added argument -Help; usage guide no longer displayed by default
#                   - NEW: added heads-up checking for unsupported commandline arguments
#                   - CHANGED: removed 'T' from 'Oldest' and 'Newest' date column values; makes table easier to read (less cluttered)
#                   - CHANGED: changed fatal error message to non-fatal warning when script's path not in Environment Path variable
#                   - FIXED: prepended script's path when importing user settings file; avoids file not in path error (thanks M Hotchin)
#                   - FIXED: fixed bug in variable $bi_folders declaration ('Aux 3e')
#                   - FIXED: fixed handling of 'Invalid session' in function BIJsonCall()
# V6   - 2023-11-06 - BREAKING CHANGE: script now gets all Blue Iris storage folder paths directly from the Windows Registry
#                   - CHANGED: eliminated user settings '$sourceN' & '$alerts_folder_path'
#                   - CHANGED: added optional user setting '$show_folderN'; $false prevents analysis of folder path #N 
#                   - CHANGED: the order in which Blue Iris folders are displayed is no longer customizable
#                   - NEW: added optional user setting '$override_folderN'; for testing; replaces folder path(s) extracted 
#                     from the Registry
#                   - NEW: script now self-checks existence of its own path in the Windows Environment System Path variable
# V5   - 2023-11-02 - BREAKING CHANGE: moved USER SETTINGS to separate script file 'BI_clip_usage_user_settings.ps1' 
#                   - NEW: added new user setting '$app_path' to customize the destination for 'output.csv' files;
#                     as a result, this script no longer outputs CSV files to 'C:\temp\bi_clip_usage'
#                   - NEW: added default time stamping of the 'output.csv' filenames, like 'output_Alerts_dat_2023-10-30_125531.csv';
#                     as a result, you can now track clip usage by folder & file type over time
#                   - NEW: added new user setting '$timestamp_output_csv_files_flg' to opt out of 'output.csv' file time stamping
#                   - NEW: added new user setting '$archive_lifetime_days' to auto-prune the time stamped 'output.csv' files
#                   - NEW: script can be executed in the background (e.g., via Windows Scheduler to save regularly save clip usage stats)
#                   - CHANGED: changed logic for determining optimum units 'TotalXX', 'AvgXX' &'MaxXX' columns
#                   - CHANGED: removed underscore characters from 'Total', 'Avg' & 'Max') column headers
# V4   - 2023-10-29 - NEW: added support for Record tab non-default filename formats, e.g., &CAM\%d-%m-%y\&CAM.%Y%m%d_%H%M%S,
#                   (thanks joshwah) also added new user setting $default_filename_format_flg
#                   - NEW: added new -DefFnameFormat argument (overrides user setting '$default_filename_format_flg')
#                   - NEW: added new user setting '$alerts_folder_path' 
#                     script no longer assumes the Alerts folder contains the substring 'Alerts'
#                   - FIXED: fixed bug resulting from $sourceN values that are naked drive letters (e.g., $source3 = 'D:\' or "D:")
#                   - CHANGED: changed folder used to store the script's temporary CSV files 
#                     from 'C:\temp\bi_clips_usage' to 'C:\temp\bi_clip_usage'
#                   - CHANGED: script no longer creates only a single temporary file 'output.csv file', 
#                     now creates a separate file for each table displayed, like 'output_Alerts_dat.csv'
#                   - NEW: added display message showing location of the script's temporary CSV files 
# V3   - 2023-10-14 - NEW: added arguments -SortColNo, -Ascending (thanks actran)
#                   - NOTE: if using '-SortColNo 1', the entry 'Total' in the 'Camera' column is tweaked to keep it at top
# V2   - 2023-10-13 - NEW: added column '#Days' to output
#                   - FIXED: fixed local time handling of dates in 'Oldest' & 'Newest' output columns
# V1   - 2023-01-05 - initial release
#
# HOW TO AUTOMATE THIS SCRIPT VIA TASK SCHEDULER
#   Open Task Scheduler and create an appropriately named task with one trigger and one action:
#   Task Scheduler
#   'Triggers' tab:
#       'At startup'
#       'Daily' or 'One time' (e.g., at 12:05 AM, every 24 hours)
#       (Note: I find 'One time' to be more reliable)
#   'Actions' tab:
#       Program/script: powershell.exe
#       Add arguments: {your_powershell_scripts_path)\{script_file_name}.ps1 args
#       EXAMPLES:
#       To run script every week
#       ... C:\ps_scripts\BI_clip_usage.ps1

param (
	[Int]$SortColNo,
	[Switch]$Ascending,
	[Switch]$DefFnameFormat,
	[Switch]$NoTimestamp,
	[Switch]$NoPruning,
	[Switch]$ClearArchive,
	[Switch]$ShowFolders,
	[Switch]$Help,
	[Switch]$DevMode
)

# For debugging
#Write-Host "SortColNo = $SortColNo"
#Write-Host "Ascending = $Ascending"

$ScriptVersion = "V8"

#Set-StrictMode  -Version 3  # https://tinyurl.com/mrx3w9v4
#Start-Transcript  # default destination is C:\Users\{user}l\Documents

# ========================================================================
# USER SETTINGS...
# Moved to separate file; see Changelog V5
# ========================================================================

# INITIALIZE VARS
$rtn_status_str = 'success'
$PathSep = [IO.Path]::DirectorySeparatorChar  # Typically '\' on Windows
$bi_folders = @(
	'New', 'Storage', 'Alerts', 
	'Aux 1', 'Aux 2', 'Aux 3', 'Aux 4', 'Aux 5', 'Aux 6', 'Aux 7', 'Aux 8', 'Aux 9', 'Aux 10', 'Aux 11', 'Aux 12', 'Aux 13'
) 

# GATHER SCRIPT & ENVIRONMENT INFO 
if (1 -eq 1) {

	#$ScriptName = $MyInvocation.MyCommand.Name
	$ScriptName = Split-Path -leaf $PSCommandpath
	$ScriptBaseName = $ScriptName.Replace('.ps1', '')
	$ScriptPath = $PSCommandPath.Replace("\$ScriptName", '')

	# For debugging
	#Write-Host "ScriptName = $ScriptName"
	#Write-Host "ScriptPath = $ScriptPath"
	#Write-Host "ScriptBaseName = $ScriptBaseName"
	#Write-Host "ScriptPath = $ScriptPath"

	$PsVersion = $PSVersionTable.PSVersion.Major
	# NOTE: Scripts running from the following hosts have process names as follows:
	#       Task Scheduler is 'svchost', Powershell-ISE is 'explorer', Windows Terminal is 'WindowsTerminal'
	#       https://stackoverflow.com/a/65894924
	$ParentProcess = (Get-Process -Id (Get-CimInstance Win32_Process -Filter "ProcessID = $pid").ParentProcessId).Name
	# ... create lookup table (used by script's CSV log)
	$HostLabels = @{}
	$HostLabels['svchost'] = 'TaskScheduler'
	$HostLabels['explorer'] = 'Powershell or Powershell-ISE'
	$HostLabels['WindowsTerminal'] = 'WindowsTerminal'

	# For testing
	#Write-Host "ParentProcess = $ParentProcess"

	# ... Get friendly host name
	if ($HostLabels[$ParentProcess]) {
		$script_host_str = $HostLabels[$ParentProcess]
	} else {
		$script_host_str = $ParentProcess
	}

	# For debugging
	#Write-Host "ParentProcess = $ParentProcess"

	# NOTE: Scripts executed in non-ISE = 'ConsoleHost', in ISE = 'Windows PowerShell ISE Host'
	$HostName = $host.name

	# For debugging
	#Write-Host "HostName = $HostName"

	# NOTE: Scripts called by another script have a 'PSCommandPath'.
	#       Method 'PSCommandPath' returns null if script is executed via the terminal
	#       https://stackoverflow.com/a/23370618/13499984
	$CallingScript = $MyInvocation.PSCommandPath

	# For testing
	#Write-Host "CallingScript = $CallingScript"

	# DETERMINE IF RUNNING IN BACKGROUND (IN CLI MODE, OR FROM TASK SCHEDULER, BLUE IRIS, etc.)
	$FromConsole = $False  # Default
	if ($null -eq $CallingScript) {
		$FromConsole = $True
	} elseif ($CallingScript -match $ScriptName) {  # Detect recursive calls
		$FromConsole = $True
	}

	# For testing
	#$fromConsole = $false

	# FEEDBACK UNEXPECTED COMMANDLINE ARGUMENTS
	#Write-Host ("params = " + ($PSBoundParameters | ConvertTo-Json))
	$cmdline_args_passed = [String[]]$PSBoundParameters.Keys
	#Write-Host "Extra Args w/ values = $Args"
	$cmdline_args_bad = ($Args | where{$_ -like "-*"}) | ForEach-Object{$_ -Replace '-', ''}
	if ($cmdline_args_bad) {
		$cmdline_args_bad = [String[]]$cmdline_args_bad
	} else {
		$cmdline_args_bad = @()
	}
	#Write-Host "Cmdline Args Passed = $cmdline_args_passed"
	#Write-Host "Cmdline Args Bad = $cmdline_args_bad"
	if ($cmdline_args_bad.Count -gt 0) {
		if ($FromConsole) {
			Write-Host ("HEADS-UP: commandline contains unsupported argument(s): [ " + ($cmdline_args_bad -Join ', ') + " ]") -ForegroundColor Green
			return
		}
		
	}
}

function get_files_recursive() {
# PURPOSE: Get file info for Blue Iris clip-type files (*.bvr, *.mp4, *.jpg, *.dat)
# RETURNS: (mixed) get_time_ms, file_names_ary, file_subdirs_ary, file_sizes_ary (bytes), $file_timestamps_ary (UNIX)
# REF: https://stackoverflow.com/questions/40220590/return-multidimensional-array-from-function
	
	param (
		[string]$Path
	)

	if ($Path -eq "") {
		return
	}

	$PathSep = [IO.Path]::DirectorySeparatorChar  # Typically '\' on Windows
	$root_path = $Path.TrimEnd($PathSep)
	
	# Initialize collectors to be returned (lists)
	$names_list   = New-Object System.Collections.Generic.List[System.Object]
	$subdirs_list = New-Object System.Collections.Generic.List[System.Object]
	$sizes_list   = New-Object System.Collections.Generic.List[System.Object]
	$dates_list   = New-Object System.Collections.Generic.List[System.Object]
	$tstamps_list = New-Object System.Collections.Generic.List[System.Object]

	$timer = [Diagnostics.Stopwatch]::StartNew()

	# GET FILTERED FILE INFO FROM SPECIFIED DIRECTORY
	# REFS: https://superuser.com/questions/1695264/getting-access-denied-when-running-get-childitem-even-as-an-administrator
	#       https://stackoverflow.com/a/12859009
	# Note: -Filter does not accept [string[]]
	$target_path = $Path + $PathSep + "*"  # like "$Path\*"
	#$files_obj = Get-ChildItem $target_path -Recurse -Filter *.bvr,*.mp4,*.jpg,*.dat -ErrorAction SilentlyContinue
	#$files_obj = Get-ChildItem $target_path -Recurse -Include *.bvr,*.mp4,*.jpg,*.dat -ErrorAction SilentlyContinue
	#$files_obj = Get-ChildItem "$Path\*" -Recurse -Include *.bvr,*.mp4,*.jpg,*.dat
	$files_obj = Get-ChildItem $target_path -Recurse -Include *.bvr,*.mp4,*.jpg,*.dat

	# For debugging
	# ... expecting type = 'System.Object[]', type name = 'Object[]'
	# Write-Host ("files_obj type = " + $files_obj.GetType().Name)

	[array]$all_filespecs_ary = $files_obj | Select -Expand FullName
	[array]$all_sizes_ary     = $files_obj | Select -Expand Length
	#[array]$all_dates_ary     = $files_obj | Select -Expand LastWriteTime
	[array]$all_dates_ary     = $files_obj | Select -Expand CreationTime | ForEach-Object { $_.ToLocalTime() }
	[array]$all_tstamps_ary   = @()  # Will be created from dates

	$timer.stop()

	if ($DevModeFlg) {
		if ($FromConsoleFlg) {
			Write-Host ("(dev mode) TOTAL folder recurse time   = " + ($timer.elapsed.totalmilliseconds).ToString("0.00") + " msec") -ForegroundColor Cyan
		}
	}

	# Parse/format file info.
	$i = 0
	foreach ($fs in $all_filespecs_ary) {

		# File names
		$name = Split-Path $fs -Leaf
		$names_list.Add($name)

		# Sub-folders; use blank for items in root directory
		$str = Split-Path $fs -Parent   # like 'D:\BlueIris2\New' or 'D:\BlueIris2\New\DW1\10-18-23'
		$parent_subfolder = $str.Replace($root_path, "")
		$subdirs_list.Add($parent_subfolder)
		
		# File sizes
		$sizes_list.Add($all_sizes_ary[$i])
		
		# File create dates
		$dt = Get-Date -Date $all_dates_ary[$i] -Format "yyyy/MM/dd"
		$dates_list.Add($dt)
		
		# File UNIX timestamps
		$ts = Get-Date -Date ($all_dates_ary[$i]).ToUniversalTime() -UFormat %s
		$tstamps_list.Add($ts)
		$i++
	}

	# Fill collectors
	$file_names_ary      = $names_list.ToArray()
	$file_subdirs_ary    = $subdirs_list.ToArray()
	$file_sizes_ary      = $sizes_list.ToArray()
	$file_dates_ary      = $dates_list.ToArray()
	$file_timestamps_ary = $tstamps_list.ToArray()

	# For debugging
	# Write-Host ("all_filespecs_ary = " + $all_filespecs_ary)
	# Write-Host ("subdirs_list = " + $subdirs_list)
	# Write-Host ("file_names_ary = " + $file_names_ary)
	# Write-Host ("file_subdirs_ary = " + $file_subdirs_ary)
	# Write-Host ("file_sizes_ary = " + $file_sizes_ary)
	# Write-Host ("file_sizes_ary type = " + $file_sizes_ary.GetType().Name)
	# Write-Host ("file_dates_ary = " + $file_dates_ary)
	# Write-Host ("file_timestamps_ary = " + $file_timestamps_ary)

	return $timer, $file_names_ary, $file_subdirs_ary, $file_sizes_ary, $file_dates_ary, $file_timestamps_ary

}

function get_column_formats($avg) {
# DEPRECATED - as of V5; started using using function get_column_formats()
# PURPOSE: For a directory fileset, use the average filesize to determine CSV column formats. 
# ARGS:    $avg = (double) 
# RETURNS: array $fact1, $fact2, $total_hdr, $avg_hdr, $max_hdr
#          where $fact1     = (double) multiplier for the total filesize column
#                $fact2     = (double) multiplier for the average and max filesize columns
#                $total_hdr = (string) column label for the total filesize column
#                $avg_hdr   = (string) column label for the average file size column
#                $max_hdr   = (string) column label for the maximum file size column
# REFS:    Powershell built-in filesize constants... https://tinyurl.com/4kn8yp3p

	# @@@ DEV - Powershell built-in filesize constants.
	#Write-Host ("---`navg = $avg")
	#Write-Host ("1TB = " + 1TB + ", (2^10)^4 = " + [Math]::Pow([Math]::Pow(2,10),4) )
	#Write-Host ("1GB = " + 1GB + ", (2^10)^3 = " + [Math]::Pow([Math]::Pow(2,10),3) )
	#Write-Host ("1MB = " + 1MB + ", (2^10)^2 = " + [Math]::Pow([Math]::Pow(2,10),2) )
	#Write-Host ("1KB = " + 1KB + ", 2^10 = " + [Math]::Pow(2,10))

	# Format for display.
	if ($avg -ge 10*1TB) {  # 10*(2^10)^4
		$fact1 = 1/1TB/1024
		$fact2 = 1/1TB
		$total_hdr = "Total_PB"
		$avg_hdr   = "Avg_TB"
		$max_hdr   = "Max_TB"
	} elseif ($avg -ge 10*1GB) {  # 10*(2^10)^3
		$fact1 = 1/1GB/1024
		$fact2 = 1/1GB
		$total_hdr = "Total_TB"
		$avg_hdr   = "Avg_GB"
		$max_hdr   = "Max_GB"
	} elseif ($avg -ge 10*1MB) {  # 10*(2^10)^2
		$fact1 = 1/1MB/1024
		$fact2 = 1/1MB
		$total_hdr = "Total_GB"
		$avg_hdr   = "Avg_MB"
		$max_hdr   = "Max_MB"
	} else {
		$fact1 = 1/1KB/1024
		$fact2 = 1/1KB
		$total_hdr = "Total_MB"
		$avg_hdr   = "Avg_KB"
		$max_hdr   = "Max_KB"
	}
	
	#Write-Host ("fact1 type = " + $fact1.GetType())

	return $fact1, $fact2, $total_hdr, $avg_hdr, $max_hdr	
}

function get_column_formats2() {
# PURPOSE: Determine CSV column formats. 
# ARGS:    -CamLabels = (unique array) of camera labels (actually '[\subfolder\]camera')) for every $type file in directory
#          -CamSizes  = (array of lists) 'sizes in bytes' for each entry in $CamLabels
# RETURNS: array $fact1, $fact2, $fact3, $total_hdr, $avg_hdr, $max_hdr
#          where $fact1     = (double) multiplier for the total filesize column
#                $fact2     = (double) multiplier for the average filesize columns
#                $fact3     = (double) multiplier for the max filesize columns
#                $total_hdr = (string) column label for the total filesize column
#                $avg_hdr   = (string) column label for the average file size column
#                $max_hdr   = (string) column label for the maximum file size column
# REFS:    Powershell built-in filesize contants... https://tinyurl.com/4kn8yp3p

	param(
		$CamLabels,  # (unique array) of camera prefixes (actually '[\subfolder\]camera')) for every $type file in directory.
		$CamSizes    # (array of lists) 'sizes in bytes' for each entry in $CamLabels
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
	
	# Write-Host ("$fn_name() CamLabels type = " + $CamLabels.GetType().Name)
	# Write-Host ("$fn_name() CamLabels = " + ($CamLabels | ConvertTo-Json))
	# Write-Host ("$fn_name() CamSizes type = " + $CamSizes.GetType().Name)
	# Write-Host ("$fn_name() CamSizes = " + ($CamSizes | ConvertTo-Json))

	# Powershell built-in filesize constants.
	<# Write-Host ("1PB = " + 1PB + ", (2^10)^5 = " + [int64]([Math]::Pow([Math]::Pow(2,10),5)) )
	Write-Host ("1TB = " + 1TB + ", (2^10)^4 = " + [Math]::Pow([Math]::Pow(2,10),4) )
	Write-Host ("1GB = " + 1GB + ", (2^10)^3 = " + [Math]::Pow([Math]::Pow(2,10),3) )
	Write-Host ("1MB = " + 1MB + ", (2^10)^2 = " + [Math]::Pow([Math]::Pow(2,10),2) )
	Write-Host ("1KB = " + 1KB + ", 2^10 = " + [Math]::Pow(2,10)) #>

	# Initialize collectors
	[array]$sums_ary = @()
	[array]$avgs_ary = @()
	[array]$maxs_ary = @()
	
	# Get stats for each camera's files (one row in the table).
	$i = 0
	foreach ($cam_lbl in $CamLabels) {
		
		[string[]]$sizes_vals = $CamSizes[$i] -Split " "

		if ($sizes_vals.Count -eq 0) {
			continue;
		}

		# Convert selected lists to value arrays.
		# REF: https://tinyurl.com/y7evp8x5
		[array]$sizes_ary = @()
		foreach ($val in $sizes_vals) {
			$sizes_ary += $val -as [Double]
		}
		
		# For debugging
		# Write-Host (""
		# Write-Host ("$fn_name() CamSizes[$i] [$cam_lbl] " + $CamSizes[$i])
		# Write-Host ($fn_name() "sizes_vals [$cam_lbl] cnt= " + $sizes_vals.Count)
		# Write-Host ("$fn_name() sizes_vals [$cam_lbl] = " + ($sizes_vals | ConvertTo-Json))
		# Write-Host ("$fn_name() sizes_ary [$cam_lbl] cnt= " + $sizes_ary.Count)
		# Write-Host ("$fn_name() sizes_ary [$cam_lbl] = " + ($sizes_ary | ConvertTo-Json))

		# Perform calculations on specified columns.
		$cnt = $sizes_ary.Count
		$sum = ($sizes_ary | Measure-Object -sum).sum  # bytes
		$max = ($sizes_ary | Measure-Object -maximum).maximum  # bytes
		$avg = ($sizes_ary | Measure-Object -average).average  # bytes
		
		# For debugging
		# Write-Host "$cam_lbl ..."
		# Write-Host "- cnt = $cnt"
		# Write-Host "- sum = $sum"
		# Write-Host "- avg = $avg"
		# Write-Host "- max = $max"
		
		# Add values to the collectors.
		$sums_ary += $sum
		$avgs_ary += $avg
		$maxs_ary += $max
		
		$i++
	}
	
	# Add 'TOTALS' row values
	$x = ($sums_ary | Measure-Object -sum).sum  # bytes
	$sums_ary += $x
	$y = ($avgs_ary | Measure-Object -average).average  # bytes
	$avgs_ary += $y
	$z = ($maxs_ary | Measure-Object -maximum).maximum  # bytes
	$maxs_ary += $z
		
	# For debugging
	#Write-Host ("sums_ary cnt= " + $sums_ary.Count)
	#Write-Host ("sums_ary = " + ($sums_ary | ConvertTo-Json))
	#Write-Host ("avgs_ary cnt= " + $avgs_ary.Count)
	#Write-Host ("avgs_ary = " + ($avgs_ary | ConvertTo-Json))
	#Write-Host ("maxs_ary cnt= " + $maxs_ary.Count)
	#Write-Host ("maxs_ary = " + ($maxs_ary | ConvertTo-Json))

	# Get the value that determines the formating for the column
	$sums_max = ($sums_ary | Measure-Object -maximum).maximum  # bytes
	$avgs_max = ($avgs_ary | Measure-Object -maximum).maximum  # bytes
	$maxs_max = ($maxs_ary | Measure-Object -maximum).maximum  # bytes

	# Format 'Total' column for display.
	if ($sums_max -ge 10*1PB) {  # 10*(2^10)^5
		$fact1 = 1/1PB
		$total_hdr = "TotalPB"
	} if ($sums_max -ge 10*1TB) {  # 10*(2^10)^4
		$fact1 = 1/1TB
		$total_hdr = "TotalTB"
	} elseif ($sums_max -ge 10*1GB) {  # 10*(2^10)^3
		$fact1 = 1/1GB
		$total_hdr = "TotalGB"
	} elseif ($sums_max -ge 10*1MB) {  # 10*(2^10)^2
		$fact1 = 1/1MB
		$total_hdr = "TotalMB"
	} else {
		$fact1 = 1/1KB
		$total_hdr = "TotalKB"
	}

	# Format 'Avg' column for display.
	if ($avgs_max -ge 10*1TB) {  # 10*(2^10)^4
		$fact2 = 1/1TB
		$avg_hdr = "AvgTB"
	} elseif ($avgs_max -ge 10*1GB) {  # 10*(2^10)^3
		$fact2 = 1/1GB
		$avg_hdr = "AvgGB"
	} elseif ($avgs_max -ge 10*1MB) {  # 10*(2^10)^2
		$fact2 = 1/1MB
		$avg_hdr = "AvgMB"
	} else {
		$fact2 = 1/1KB
		$avg_hdr = "AvgKB"
	}

	# Format 'Max' column for display.
	if ($maxs_max -ge 10*1TB) {  # 10*(2^10)^4
		$fact3 = 1/1TB
		$max_hdr = "MaxTB"
	} elseif ($maxs_max -ge 10*1GB) {  # 10*(2^10)^3
		$fact3 = 1/1GB
		$max_hdr = "MaxGB"
	} elseif ($maxs_max -ge 10*1MB) {  # 10*(2^10)^2
		$fact3 = 1/1MB
		$max_hdr = "MaxMB"
	} else {
		$fact3 = 1/1KB
		$max_hdr = "MaxKB"
	}
	
	# For debugging
	#Write-Host "$total_hdr fact1 = $fact1"
	#Write-Host "$max_hdr   fact2 = $fact2"
	#Write-Host "$max_hdr   fact3 = $fact3"

	return $fact1, $fact2, $fact3, $total_hdr, $avg_hdr, $max_hdr	
}

function show_table($type, $folder, $sortcolno, $ascendingflg, $files_cams, $files_lists, $sizes_lists, $tstamps_lists, $days_ary, $all_sizes_ary, $all_tstamps_ary, $OutputPath, $TimestampOutputFlg, $FromConsoleFlg, $DevModeFlg) {
# PURPOSE:  Output pretty table of 'clips' of the specified $type extracted from a single Blue Iris clips directory.
#           Do this using a temporary CSV file; see note 3 below.
# ARGS:     $type               = (string) file type (extension) 'bvr', 'mp4', 'jpg', OR 'dat'
#           $folder             = (string) Blue Iris folder name, like 'D:\BlueIris\New|Stored|Alerts', 'D:'...
#           $sortcolno          = (int) sort column number (1 to N)
#           $ascendingflg       = (boolean) sort column number ascending (default is descending)
#           $files_cams         = (array) unique array of camera prefixes (actually '[\subfolder\]camera'))for every $type file in directory.
#           $files_lists        = (array of lists) '[\subfolder\]filenames' for each entry in $files_cams
#           $sizes_lists        = (array of lists) 'sizes in bytes' for each entry in $files_cams
#           $tstamp_lists       = (array of lists) 'UNIX timestamps' for each entry in $files_cams
#           $days_ary           = (array) number of days for every $type file in directory.
#           $all_sizes_ary      = (array) 'sizes in bytes' for every $type file in directory.
#           $all_tstamps_ary    = (array) 'UNIX timestamps' for every $type file in directory.
#           $OutputPath         = (string) full path to folder for 'output.csv' files
#           $TimestampOutputFlg = (boolean) prefix output.csv files with timestamp, like '2023-10-29_1215'
#           $FromConsoleFlg     = (boolean) $true notifies function that calling command was executed from the console
#           $DevModeFlg         = (boolean)
# RETURNS:  $timer1, $timer2
# REQUIRES: get_column_formats2(), 
# NOTES:  1. ARGS use a 'array of lists' approach b/c I was unable to successfully pass a hashtable directly to this function.
#         2. Clips must use Blue Iris' default file naming format beginning w/ '&CAM'). Other patterns are *not* supported.
#         3. Using CSV files is very convenient; Powershell automatically handles the table formatting.
#         4. The temporary CSV files displayed by this utility are generated using the following command:
#            --> Import-Csv "output.csv" | Sort-Object $total_hdr -Descending | Format-Table 
#            A challenge with this approach is that the Sort-Object column has only the -Length property.
#            Because of this, successfully displaying the total bytes column in descending order requires that its values  
#            i) have an identical number of digits after the decimal, and 
#            ii) be left-padded to create strings of identical length.
#         5. The temporary CSV files are created in directory $app_path.

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	# Write-Host "Entered function: $fn_name() ..."	
	# Write-Host ("$fn_name() type = " + $type)
	# Write-Host ("$fn_name() folder = " + $folder)
	# Write-Host ("$fn_name() sortcolno = " + $sortcolno)
	# Write-Host ("$fn_name() ascendingflg = " + $ascendingflg)
	# if ($folder.indexOf("Aux1") -gt -1) {
		# Write-Host ("$fn_name() files_cams type = " + $files_cams.GetType().Name)
		# Write-Host ("$fn_name() files_cams = " + $files_cams)
		# Write-Host ("$fn_name() files_cams keys cnt = " + $files_cams.length)
		# Write-Host ("$fn_name() files_lists type = " + $files_lists.GetType().Name)
		# Write-Host ("$fn_name() files_lists cnt = " + $files_lists.length)
		# Write-Host ("$fn_name() sizes_lists type = " + $sizes_lists.GetType().Name)
		# Write-Host ("$fn_name() sizes_lists cnt = " + $sizes_lists.length)
		# Write-Host ("$fn_name() sizes_lists = " + $sizes_lists)
		# Write-Host ("$fn_name() days_ary = " + $days_ary)
		# Write-Host ("$fn_name() days_ary cnt = " + $days_ary.length)
		# Write-Host ("$fn_name() tstamps_lists cnt = " + $tstamps_lists.length)
		# Write-Host ("$fn_name() tstamps_lists = " + $tstamps_lists)
		# Write-Host ("$fn_name() all_sizes_ary type = " + $all_sizes_ary.GetType())
		# Write-Host ("$fn_name() all_sizes_ary cnt = " + $all_sizes_ary.Length)
		# Write-Host ("$fn_name() all_tstamps_ary type = " + $all_tstamps_ary.GetType())
		# Write-Host ("$fn_name() all_tstamps_ary len = " + $all_tstamps_ary.Length)
		# Write-Host ("$fn_name() OutputPath = " + $OutputPath)
		# Write-Host ("$fn_name() TimestampOutputFlg = " + $TimestampOutputFlg)
		# Write-Host ("$fn_name() FromConsoleFlg = " + $FromConsoleFlg)
		# Write-Host ("$fn_name() DevModeFlg = '" + $DevModeFlg + "'")
	# }

	$PathSep = [IO.Path]::DirectorySeparatorChar  # Typically '\' on Windows

	if ($folder.EndsWith($PathSep) ){
		$folder = $folder.TrimEnd($PathSep)
	}
		
	if ($TimestampOutputFlg) {
		$ts_str = Get-Date -Format "yyyy-MM-dd_HHmmss"
		$ts_str = "_" + $ts_str
	} else {
		$ts_str = ""
	}

	# Initialize list collectors - each builds a column in the displayed output table.
	$cams  = New-Object System.Collections.Generic.List[System.Object]
	$types = New-Object System.Collections.Generic.List[System.Object]
	$cnts  = New-Object System.Collections.Generic.List[System.Object]
	$szs   = New-Object System.Collections.Generic.List[System.Object]
	$avgs  = New-Object System.Collections.Generic.List[System.Object]
	$maxs  = New-Object System.Collections.Generic.List[System.Object]
	$olds  = New-Object System.Collections.Generic.List[System.Object]
	$news  = New-Object System.Collections.Generic.List[System.Object]
	$days  = New-Object System.Collections.Generic.List[System.Object]

	# Get system's timezone info
	$tzInfo = [System.TimeZoneInfo]::Local  # System's timezone info

	$timer1 = [Diagnostics.Stopwatch]::StartNew()

	# Get stats for entire file set.
	$avg_all = ($all_sizes_ary | Measure-Object -average).average  # bytes
	$max_all = ($all_sizes_ary | Measure-Object -maximum).maximum  # bytes
	$old_all = ($all_tstamps_ary | Measure-Object -minimum).minimum  # UNIX timestamp, seconds
	$new_all = ($all_tstamps_ary | Measure-Object -maximum).maximum  # UNIX timestamp, seconds
	
	$timer1.stop()
	
	if ($avg_all -gt 0) {
		
		if ($DevModeFlg) {
			if ($FromConsoleFlg) {
				Write-Host ("(dev mode) TOTAL row stats calc time   = " + ($timer1.elapsed.totalmilliseconds).ToString("0.00") + " msec") -ForegroundColor Cyan
			}
		}

		# Size the filesize columns ('Size', 'Avg', 'Max')
		# ... based on the average for the entire collection
		$fact1, $fact2, $fact3, $total_hdr, $avg_hdr, $max_hdr = `
			get_column_formats2 `
				-CamLabels $files_cams `
				-CamSizes $sizes_lists  
		
		$timer2 = [Diagnostics.Stopwatch]::StartNew()

		# GET STATS FOR EACH CAMERA'S FILES (ONE ROW IN THE TABLE).
		$i = 0
		foreach ($fcam in $files_cams) {
			
			# Convert selected lists to value arrays.
			# REF: https://tinyurl.com/y7evp8x5
			[string[]]$sizes_vals = $sizes_lists[$i] -split " "
			$j = 0
			foreach ($val in $sizes_vals) {
				$sizes_vals[$j] = $val -as [Double]
				$j++
			}
			[string[]]$tstamps_vals = $tstamps_lists[$i] -split " "
			$j = 0
			foreach ($val in $tstamps_vals) {
				$tstamps_vals[$j] = $val -as [Int]
				$j++
			}

			# Perform calculations on specified columns.
			$cnt = $sizes_vals.Length
			$sum = ($sizes_vals | Measure-Object -sum).sum  # bytes
			$avg = ($sizes_vals | Measure-Object -average).average  # bytes
			$max = ($sizes_vals | Measure-Object -maximum).maximum  # bytes
			$old = ($tstamps_vals | Measure-Object -minimum ).minimum
			$new = ($tstamps_vals | Measure-Object -maximum ).maximum
			
			# For debugging
			# if ($folder.indexOf("Aux1") -gt -1) {
				# Write-Host "$fcam ..."
				# Write-Host "- cnt = $cnt"
				# Write-Host "- sum = $sum"
				# Write-Host "- avg = $avg"
				# Write-Host "- max = $max"
				# Write-Host "- old = $old"
				# Write-Host "- new = $old"
			# }

			# Format column entries for display.
			# ... all numbers entries must have an identical number of digits after the decimal
			# ... see NOTE #4 above.
			$sz  = ('{0:0.#0}' -f ($sum*$fact1))
			$avg = ('{0:0.#0}' -f ($avg*$fact2))
			$max = ('{0:0.#0}' -f ($max*$fact3))

			$dt = [System.DateTimeOffset]::FromUnixTimeSeconds($old).DateTime.ToLocalTime()  # Rounds up seconds
			$isDST = $tzInfo.IsDaylightSavingTime($dt)
			# ... get the correct full timezone name, like 'Pacific Standard Time' or 'Pacific Standard Time'
			$tzName = if ($isDST) { $tzInfo.DaylightName } else { $tzInfo.StandardName }
			# ... convert long timezone name to short abbreviation (works in both PS5.1 and PS7+)
			#     use regex to extract the first letter of each word 
			#     (I prefer this to using a mapping table, which may require maintenance)
			$tzAbbreviation = ($tzName -creplace '\b(\w)\w*', '$1') -replace '\s+',''
			$dt_str_old = $dt.ToString('yyyy-MM-dd HH:mm') + " $tzAbbreviation"

			$dt = [System.DateTimeOffset]::FromUnixTimeSeconds($new).DateTime.ToLocalTime()  # Rounds up seconds
			$isDST = $tzInfo.IsDaylightSavingTime($dt)
			$tzName = if ($isDST) { $tzInfo.DaylightName } else { $tzInfo.StandardName }
			$tzAbbreviation = ($tzName -creplace '\b(\w)\w*', '$1') -replace '\s+',''
			$dt_str_new = $dt.ToString('yyyy-MM-dd HH:mm') + " $tzAbbreviation"
			
			# Add formated values to the collectors.
			if ($sum -gt 0) {
				$cams.Add($fcam)
				$types.Add($type)
				$cnts.Add($cnt)
				$szs.Add($sz)
				$avgs.Add($avg)
				$maxs.Add($max)
				$olds.Add($dt_str_old)
				$news.Add($dt_str_new)
				$days.Add($days_ary[$i])
			}
			
			$i++
		}
				
		$timer2.stop()

		if ($DevModeFlg) {
			if ($FromConsoleFlg) {
				Write-Host ("(dev mode) camera rows stats calc time = " + ($timer2.elapsed.totalmilliseconds).ToString("0.00") + " msec") -ForegroundColor Cyan
			}
		}

		# For debugging
		#Write-Host ("cams cnt = " + $cams.Count)
		
		# INSERT AN 'Total' ROW IN THE TABLE.
		# ... Note: it will appear at the top of the table after a descending sort on the total size column.
		if ($cams.Count -gt 1) {
			
			$cnts_total  = ($cnts | Measure-Object -sum ).sum
			$sizes_total = ($szs | Measure-Object -sum ).sum
			
			# Kludge to get the TOTAL row to appear at the top of the table when sorting on Column 1
			if ($sortcolno -eq 1) { 
				if ($ascendingflg) {
					$cams.Add("_TOTAL")
				} else {
					# NOTE: The following does not work in PS5
					# Prefix header with UNCODE character for symbol 'pi' (always sorts after Z)
					# REF: https://stackoverflow.com/questions/17264717/simple-to-enter-unicode-character-that-would-sort-after-z-in-most-cases
					#$cams.Add("`u{03C0}TOTAL") 
					#$cams.Add([char]0x03C0 + "TOTAL")

					$cams.Add("zTOTAL") 
				}
			} else {
				# Typical case
				$cams.Add("TOTAL")
			}
			
			$types.Add($type)
			$cnts.Add($cnts_total)
			$szs.Add($sizes_total.ToString("0.00"))
			#$avgs.Add([Math]::Round($avg_all*$fact2, 2))
			$avgs.Add(('{0:0.#0}' -f ($avg_all*$fact2)))
			#$maxs.Add([Math]::Round($max_all*$fact3, 2))
			$maxs.Add(('{0:0.#0}' -f ($max_all*$fact3)))
			$days.Add('')
			$olds.Add('')
			# ... See 'UNIX timestamp conversion' comments above regarding formatting
			#$date_str = (([System.DateTimeOffset]::FromUnixTimeSeconds($old_all)).DateTime.ToLocalTime()).ToString('s')
			#$olds.Add($date_str.Replace('T',' '))
			$news.Add('')
			#$date_str = (([System.DateTimeOffset]::FromUnixTimeSeconds($new_all)).DateTime.ToLocalTime()).ToString('s')
			#$news.Add($date_str.Replace('T',' '))
			
			# For debugging
			#Write-Host ("System.DateTimeOffset = " + ([System.DateTimeOffset]::Now))
			#Write-Host ("datetimeoffset = " + ([System.datetimeoffset]::Now))
		}
		
		# Convert all numeric value columns ('Count', 'Size', 'Avg', 'Max', 'Days') to left-padded strings.
		# ... for a pretty table, these left-padded strings must be of identical length.
		# ... see also NOTE #4 above.
		$i = 0
		[array]$arr = foreach($val in $cnts) { ([string]$val) }
		$max_len = ($arr | Measure-Object -Maximum -Property Length).Maximum
		if ($max_len -le 5) { $max_len = 5 }  # Header 'Count'
		foreach ($item in $cnts.ToArray()) {
			$val = $item.toString()
			$cnts[$i] = $val.padLeft($max_len," ")
			$i++
		}

		$i = 0
		[array]$arr = foreach($val in $szs) { ([string]$val) }
		$max_len = ($arr | Measure-Object -Maximum -Property Length).Maximum
		$hdr_len = $total_hdr.length
		if ($max_len -le $hdr_len) { $max_len = $hdr_len } 
		foreach ($item in $szs.ToArray()) {
			$val = ([decimal]$item).toString("0.00")
			$szs[$i] = $val.padLeft($max_len," ")
			$i++
		}
		#$total_hdr = $total_hdr.padRight($max_len," ")  # NO! (see BEST PRACTICES below)

		$i = 0
		[array]$arr = foreach($val in $avgs) { ([string]$val) }
		$max_len = ($arr | Measure-Object -Maximum -Property Length).Maximum
		$hdr_len = $avg_hdr.length
		if ($max_len -le $hdr_len) { $max_len = $hdr_len } 
		foreach ($item in $avgs.ToArray()) {
			$val = ([decimal]$item).toString("0.00")
			$avgs[$i] = $val.padLeft($max_len," ")
			$i++
		}
		#$avg_hdr = $avg_hdr.padRight($max_len," ")  # NO! (see BEST PRACTICES below)

		$i = 0
		[array]$arr = foreach($val in $maxs) { ([string]$val) }
		$max_len = ($arr | Measure-Object -Maximum -Property Length).Maximum
		$hdr_len = $max_hdr.length
		if ($max_len -le $hdr_len) { $max_len = $hdr_len } 
		foreach ($item in $maxs.ToArray()) {
			#$val = '{0:N2}' -f [decimal]$item
			$val = ([decimal]$item).toString("0.00")
			$maxs[$i] = $val.padLeft($max_len," ")
			$i++
		}
		#$max_hdr = $max_hdr.padRight($max_len," ")  # NO! (see BEST PRACTICES below)

		$i = 0
		[array]$arr = foreach($val in $days) { ([string]$val) }
		$max_len = ($arr | Measure-Object -Maximum -Property Length).Maximum
		$hdr_len = "#Days".length
		if ($max_len -le $hdr_len) { $max_len = $hdr_len } 
		foreach ($item in $days.ToArray()) {
			$val = $item.toString()
			$days[$i] = $val.padLeft($max_len," ")
			$i++
		}
		
		# Create & populate a hashtable with formated entries.
		# ... each element will be a column in the final table.
		# ... a hashtable is required for creating the temporary CSV file.
		$output = [ordered]@{}
		
		# Populate header row labels
		# BEST PRACTICES/TIPS: 
		#  1. Use only letters and numbers; do not use spaces and special characters (like underscores, slashes, parentheses, brackets, etc.) **
		#  2. This includes padding header labels with spaces in an attempt to improve header and data alignment ***
		#  3. To improve readability, use 'caseSeparation' or 'CamelCase' styles.
		#  **  The '#' (number sign) character is okay 
		#  *** The screen output will look ok, but the CSV file will not; the column's data will be blank! 
		$output.Add("Camera", $cams)
		$output.Add("Type", $types)
		$output.Add("Count", $cnts)
		$output.Add($total_hdr, $szs)
		$output.Add($avg_hdr, $avgs)
		$output.Add($max_hdr, $maxs)
		#$output.Add("Days", $days)
		$output.Add("#Days", $days)
		$output.Add("Oldest", $olds)
		$output.Add("Newest", $news)

		# USE A TEMPORARY CSV FILE TO DISPLAY A NICE LOOKING TABLE.
		if ($cams.Count -gt 0) {

			# EXPORT THE HASHTABLE AS A CSV FILE.
			# REFS: 1. Microsoft everything about hashtables...https://tinyurl.com/3eezjed9
			#       2. https://tommymaynard.com/hash-table-to-csv-revisited/
			#       3. PSObject vs PSCustomObject... https://stackoverflow.com/a/50196654/13499984

			# Get the maximum number of entries ($cnt) from all the columns in the CSV table.
			# ... Note: calculating this is *not* necessary in our case, b/c we already know that all
			#     the columns have the same number of entries; however, it is a good practice in general.
			$cnt = $output.Values |
				ForEach-Object { $_.Count } |
				Sort-Object |
				Select-Object -Last 1

			# Get the CSV table column headers.
			$keys = $output.Keys
			
			# For debugging
			#Write-Host ("keys = " + ($keys | ConvertTo-Json))
			
			# Must create array to use $keys in PS5
			# $keys type = System.Collections.Specialized.OrderedDictionary+OrderedDictionaryKeyValueCollection
			#$keys
			$headers_ary = @()
			foreach ($key in $keys) {
				$headers_ary += $key
			}
		
			# For debugging
			#Write-Host ("keys (column headers) = " + $keys)
			#Write-Host ("keys type = " + $keys.GetType())
			#Write-Host ("headers_ary = " + $headers_ary)
			#Write-Host ("column cnts = " + ($output.Values | ForEach-Object { $_.Count } ))
			
			# Fix sort column number if out of range
			if ($sortcolno -gt 0) {
				if ($sortcolno -gt $keys.Count) {
					$sortcolno = 0
				}
			}
			
			# Make CSV filespec
			# ... Handle drive letter only folders, like 'E:"
			if ($folder.EndsWith(':')) {
				$arr = $folder -Split ':'
				$folder = $arr[0]

				# For debugging
				#Write-Host "folder mod = $folder"

			}
			$csv_filespec = "$OutputPath\" + "output_" + $folder + "_" + $type + $ts_str + ".csv"
			
			# For debugging
			#Write-Host "csv_filespec = $csv_filespec"

			# Use hashtable $output to create a custom object for exporting.
			# ... this is necessary to save it correctly to CSV.
			# ... Export-Csv works with objects. It expects properties and values.
			# Build CSV file, row-by-row. 
			0..($cnt-1) | ForEach-Object {
				$tmp_tbl = [ordered]@{}
				foreach ($key in $keys) {
					
					# For debugging
					#Write-Host ("key = $key; `$_ = " + $_)
					#Write-Host ("output[$key] = " + ($output[$key] | ConvertTo-Json))

					$tmp_tbl[$key] = $output[$key][$_]
				}
				New-Object -Type PSObject -Property $tmp_tbl
			} | Export-Csv "$csv_filespec" -NoTypeInformation
			
			# For deb ugging
			#Write-Host ("tmp_tbl = " + ($tmp_tbl | ConvertTo-Json))

			# DISPLAY THE TEMPORARY CSV FILE, sorting on the Total Size column.
			if ($FromConsoleFlg) {
				if ($sortcolno -eq 0) {
					Import-Csv "$csv_filespec" | Sort-Object $total_hdr -Descending | Format-Table 
				} else {
					#$sort_col_hdr = $keys[$sortcolno - 1]
					$sort_col_hdr = $headers_ary[[Int]$sortcolno - 1]
			
					# For debugging
					#Write-Host ("sort_col_hdr = " + $sort_col_hdr)
					
					if ($ascendingflg) {
						Import-Csv "$csv_filespec" | Sort-Object $sort_col_hdr  | Format-Table 
					} else {
						Import-Csv "$csv_filespec" | Sort-Object $sort_col_hdr -Descending | Format-Table 					
					}
				}
				Write-Host "Note: The table above has been saved to '$csv_filespec'"
			}				
		}
	}
}

Function Convert-FromUnixDate ($UnixDate) {
   [timezone]::CurrentTimeZone.ToLocalTime(([datetime]'1/1/1970').AddSeconds($UnixDate))
}

function PurgeOldFiles() {
# PURPOSE:  Delete old files (> N days) from a filepath like 'D:\myfolder\*_20*.csv'
# ARGS:     see param lines below
# RETURNS:  (array) @(found_files_cnt, purged_files_cnt, oldest_file_path, oldest_file_lastwritetime, oldest_file_age_days)
#           or (string) 'ERR/' message
# ATTRIBUTIONS:  https://stackoverflow.com/a/46086926
#
# V1 (01/29/2025) - modified line filling var $cur_ary to sort the resutl oldest to newest
	
    param(
		[string]$Pattern,  # [REQUIRED] full filespec pattern for targeted files; like 'D:\BI_clip_usage\*_20*.csv"'
		[Int]$Days         # [REQUIRED] maximum age for retained files; use 24 for 1 day, 168 for 1 week... 
	)
	
	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function = $fn_name()"
	#Write-Host "$fn_name() Pattern = $Pattern"
	#Write-Host "$fn_name() Days = $Days"
	
	if ($null -eq $Pattern -or $null -eq $Days) {
		Write-Host ("ERR/$fn_name() requires 2 arguments (-Pattern -Days)") -ForegroundColor Red
		return
	}
	
	# GET FILES MATCHING PATTERN, sorted oldest to newest
	$cur_ary = @(Get-ChildItem -Path "$Pattern" | Sort-Object LastWriteTime | Select-Object Name) # array of hashtables

	if ($cur_ary) {
		$cur_ary_cnt = $cur_ary.Count
	} else {
		$cur_ary_cnt = 0
	}

	# For debugging
	# Write-Host ("$fn_name() cur_ary type = " + $cur_ary.GetType())
	# Write-Host "$fn_name() cur_ary cnt = $cur_ary_cnt"
	# Write-Host ("$fn_name() cur_ary[0] = " + $cur_ary[0])
	# Write-Host ("$fn_name() cur_ary[" + ($cur_ary_cnt - 1) + "] = " + $cur_ary[$cur_ary_cnt - 1])
	# Write-Host ("$fn_name() cur_ary = " + ($cur_ary | ConvertTo-Json))
	
	# Oldest file age in days
	$PathSep = [IO.Path]::DirectorySeparatorChar  # Typically '\' on Windows
	$path = Split-Path $Pattern -Parent
	$oldest_file_path = $path + $PathSep + $cur_ary[0].Name
	$file_obj = Get-Item -Path $oldest_file_path
	$oldest_file_dt = $file_obj.LastWriteTime
	$oldest_file_days = (Get-Date) - $file_obj.LastWriteTime | Select-Object -ExpandProperty Days

	# For debugging
	# Write-Host ("oldest file = '" + $oldest_file_path + "'")
	# Write-Host ("oldest file datetime = " + $oldest_file_dt)
	# Write-Host ("oldest file age in days = " + $oldest_file_days)	

	# GET FILES MATCHING AGE REQUIREMENT
	$purge_ary = @() 		
	#if ($cur_ary_cnt -gt 0 -and $Days -gt 0) {
	if ($cur_ary_cnt -gt 0 -and $Days -ne 0) {  #  suggested by M Hotchin 5/16/2024
		
		$limit = (Get-Date).AddDays(-$Days)
		
		# For debugging
		#Write-Host "$fn_name() limit = $limit"

		$purge_obj = Get-ChildItem -Path "$Pattern" | 
			Where-Object { !$_.PSIsContainer -and $_.LastWriteTime -lt $limit } | 
			Select Name

		if ($purge_obj) {
			$purge_ary = $purge_obj.Name
		}
	}
	
	# Get count for return array
	$purge_ary_cnt = $purge_ary.Count
	
	# For debugging
	#Write-Host "$fn_name() purge_ary cnt = $purge_ary"
	#Write-Host ("$fn_name() purge_ary = " + ($purge_ary | ConvertTo-Json))

	# REMOVE FILES MATCHING AGE REQUIREMENT
	if ($purge_ary_cnt -gt 0) {
		if ($Days) {
			$null = Get-ChildItem -Path $Pattern | 
				Where-Object { !$_.PSIsContainer -and $_.LastWriteTime -lt $limit } | 
				Remove-Item -Force
		}
	}
	
	return  @($cur_ary_cnt, $purge_ary.Count, $oldest_file_path, $oldest_file_dt, $oldest_file_days)
}

function RegGetValue() {
# PURPOSE:  Get current value of a registry entry
# ARGS:     see param lines below
# RETURNS:  registry value, user-provided default value, or error message
# 
# V2 (2024/07/01)
# Created by jaydeel @ ipcamtalk.com

	param(
		[String]$KeyPath,  # like, 'HKLM:SOFTWARE\Perspective Software\Blue Iris\clips'
		[String]$KeyName,  # like, 'compact'
		$DefaultValue      # value to return if $KeyName does not exist 
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
	
	# For debugging
	#Write-Host "Entered $fn_name() ..."
	#Write-Host "$fn_name() -KeyPath = '$KeyPath'"
	#Write-Host "$fn_name() -KeyName = '$KeyName'"
	#Write-Host "$fn_name() -DefaultValue = '$DefaultValue'"


	try {
		$reg_obj = Get-ItemProperty -Path $KeyPath -Name $KeyName -ErrorAction Stop
	
		# For debugging
		#Write-Host ("reg_obj = " + ($reg_obj | ConvertTo-Json -Depth 100))
		
		if ($reg_obj) {
			if (!($null -eq $reg_obj.$KeyName)) {
				return $reg_obj.$KeyName
			}
		}
	}
	catch {
		# continue
	}	

	if ($DefaultValue) {
		$rtn_value = $DefaultValue
	} else {
		$err_msg = "ERR/unable to get registry key '$KeyPath\$KeyName'"
		$rtn_value = $err_msg
	}
	
	return $rtn_value
}

function BIJsonCall() {
# PURPOSE:   To get a Blue Iris JSON interface response from the Blue Iris web server.
#            Intended to be called by other PowerShell scripts.
# ARGUMENTS: -JsonCmd        = (hashtable) like '@{cmd=xxx},@{cmd=xxx,camera=name},...'
#            -ForceFlg       = [OPTIONAL] (boolean) forces new login (gets a new authenticated session id)
#            -RetryNo        = [OPTIONAL] (int) if -gt 0, this call is a retry attempt
#            -QuietFlg       = [OPTIONAL] (boolean) $true suppresses verbose output to the console
#            -FromConsoleFlg = [OPTIONAL] (boolean) $true notifies function that calling command was executed from the console
# RETURNS:   (object) PowerShell PSCustomObject
# REFS:      1) Blue Iris Help PDF: chapter 'Administration', section 'JSON Interface'
#            2) https://www.houselogix.com/docs/blue-iris/BlueIris/json.htm  (old!)
#            3) https://ipcamtalk.com/threads/json-interface.21491/
#            4) https://ipcamtalk.com/threads/json-help-please.40381/#post-372770
# NOTES:
# 1) Saves the BI API session id to a file for reuse (to avoid unnecessary repeated 
#    logins while the session is still active); the file stored in folder 
#    'c:\temp\$ScriptBaseName' (created if necessary)
# 2) In the event that the JSON call fails with reason 'Invalid session',
#    the script will automatically attempt to repeat the call up to $max_retries
#
# Version: 3 (2024-02-17)
# Created by jaydeel at ipcamtalk.com

	param (
		[Hashtable]$JsonCmd,
		[Boolean]$ForceFlg,
		[Int]$RetryNo,
		[Boolean]$QuietFlg,
		[Boolean]$FromConsoleFlg
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered function: $fn_name() ..."	
	#Write-Host "`Entered function $fn_name()`n-----------------------------"
	#Write-Host ("$fn_name() JsonCmd = " + ($JsonCmd | ConvertTo-Json))
	#Write-Host "$fn_name() PsVersion = $PsVersion"	
	#Write-Host "$fn_name() FromConsoleFlg = $FromConsoleFlg"

	# For debugging
	#Write-Host "$fn_name() retry_cntr = $retry_cntr"	

	# System vars
	$ScriptName = Split-Path -leaf $PSCommandpath
	$ScriptBaseName = $ScriptName.Replace('.ps1', '')
	$ParentProcess = (Get-Process -Id (Get-CimInstance Win32_Process -Filter "ProcessID = $pid").ParentProcessId).Name
	$PsVersion = $PSVersionTable.PSVersion.Major
	# Get the current server domain (hostname) and username
	# Used for background messaging via msg.exe
	# ... tip: from the console, you can type 'hostname' or 'whoami'
	$server_name = (whoami).Split('\')[0]
	$server_username = (whoami).Split('\')[1]

	function encryptMD5($str) {
	# PURPOSE: Get MD5 hash for a string.
	# ARGS:    $str = (string) string to be encrypted
	# RETURNS: (string) MD5 encrypted string... like, '0753a0e20c9027ccab70f4291a21aa08'
	# REF:     https://infosecscout.com/get-md5-hash-in-powershell/
	# NOTES:   (from REF)
	#	Unfortunately, PowerShell provides no direct function to generate a hash from a string.
	#	However, PowerShell offers a cmdlet to generate MD5 hash for a file: Get-FileHash. 
	# 	The cmdlet can also be used to get the MD5 hash for a string, by opening a stream and hashing it.
	# UTILITY: https://www.md5hashgenerator.com/
	# ... Example: 'MD5string' returns '0753a0e20c9027ccab70f4291a21aa08'

		$stringAsStream = [System.IO.MemoryStream]::new()
		$writer = [System.IO.StreamWriter]::new($stringAsStream)
		$writer.write($str)
		$writer.Flush()
		$stringAsStream.Position = 0
		$hash_obj = Get-FileHash -InputStream $stringAsStream -Algorithm MD5
		
		# For debugging
		#$str = $hash_obj | Out-String
		#Write-Host "hash_obj = $str"

		return $hash_obj.Hash
	}
	
	# MAX RETRIES FOR JSON CMD
	# ... This function will recursively call itself if a JSON cmd's previous
	#     $json_response.result value is 'fail' and the reason is 'Invalid session'
	$max_retries = 3
	# ... Wait time before attempting a 'Invalid session' recursive call 
	$retry_delay_ms = 100
	
	if (!$RetryNo) {
		$retry_cntr = 0
	} else {
		$retry_cntr = $RetryNo
	}

	# Abort if we've exceeded the retries 'Invalid session' break variable
	if ($retry_cntr -ge $max_retries) {
		$err_msg ="ERR/JsonCmd '" + $JsonCmd.cmd + "' unable to create a valid session"
		return $err_msg
	}
		
	# Show progress bar as a heads-up visualization that retries are being attempted
	<# if ($retry_cntr -gt 0) {
		if ($fromConsole) {
			$WriteProgressParams = @{
				Activity         = "Retrying BI JSON Call..."
				Status           = "Progress: retry $retry_cntr of $max_retries max"
				PercentComplete  = (($retry_cntr / $max_retries) * 100)
			}
			Write-Progress @WriteProgressParams
			# Note: If loop is too fast, you might need a delay to get the progress bar to show
			Start-Sleep -Milliseconds 2000				
		}
	} #>

	# HANDLE FILES FOR SAVED SESSION IDS AND RETRY COUNTER
	# ... Perform existence test for the *.txt file used to save session ids
	#     Create its path if it does not exist
	#     NOTE: if a path is not specified, the default Windows location is 'C:\Users\<username>"
	$path = "c:\temp\$ScriptBaseName"
	if (!(Test-Path -LiteralPath $path)) {
		New-Item -ItemType Directory -Force -Path $path
	}
	# ... NOTE: we use a different filename when running the script via different processes;
	#     this avoids collisions with scripts running via:
	#     'scvhost' (Task Scheduler), 'WindowsTerminal', 'explorer' (Powershell ISE)
	$base_name = "$path\bi_session_id"
	$session_filespec = $base_name + "_" + $ParentProcess + ".txt"
	$found_saved_session_file_flg = (Test-Path -LiteralPath $session_filespec)

	# PARSE COMMANDLINE ARGS
	if ($JsonCmd -eq $null) {
		$err_msg = "ERR/expecting an argument: JsonCmd (PowerShell hashtable)"
		return $err_msg
	} else {
		# For debugging
		#Write-Host ("requested JsonCmd type = " + ($JsonCmd.GetType()))
	}
	# ... Save original JsonCmd (in case we need to make a recursive call)
	$json_cmd_orig = $JsonCmd
	# ... Override reuse of saved session_id
	if ($ForceFlg) {
		$found_saved_session_file_flg = $false
	}

	# ========================
	# PERFORM API CALLS

	# Make the URL
	$uri = "http://" + $script:bi_webserver + "/json";

	# For debugging
	#Write-Host "$fn_name() uri = $uri"

	# Flag 'logout' JSON cmd requests
	# ... NOTE: All JSON 'logout' cmds should be performed by the calling script!
	if ($JsonCmd.cmd -eq 'logout') {
		$logging_out_flg = $true
	} else {
		$logging_out_flg = $false
	}
	
	# For debugging
	if ($logging_out_flg) {
		#Write-Host "$fn_name() detected JSON cmd to log out of BI API session"
	} else {
		#Write-Host "$fn_name() found_saved_session_file_flg... $found_saved_session_file_flg"
		#msg.exe * /TIME:5 "$fn_name() found_saved_session_file_flg... $found_saved_session_file_flg"
	}

	# SPECIAL HANDLING FOR 'geofence' JSON COMMAND REQUESTS
	$dev_uuid = $null  # Default
	$json_cmd_keys = $JsonCmd.Keys
	if ($json_cmd_keys -contains 'uuid') {
		if ($JsonCmd.uuid -ne $null) {
			$dev_uuid = $JsonCmd.uuid
		}
	}

	# For debugging
	if ($dev_uuid) {
		#Write-Host "$fn_name() dev_uuid = $dev_uuid"
	}

	# EXECUTE API CALL
	
	if (!$found_saved_session_file_flg) {
		
		# ================================================================
		# 1.1 PRELIMINARY LOGIN
		
		# Create, then encode the JSON 'login' command
		$parameters = @{
			cmd = 'login'
		}
		$json_params = $parameters | ConvertTo-Json

		# For debugging
		#Write-host "$fn_name() (1.1) login #1 json_params... `n$json_params"
		#msg.exe $server_username /TIME:15 "$fn_name() (1.1) login #1 json_params... `n$json_params"

		# Execute the JSON request
		$jsonResponseObj = $json_params | Invoke-RestMethod -Uri $uri -Method Post

		# For debugging
		#msg.exe $server_username /TIME:15 "$fn_name() (1.1) login #1 jsonResponse... `n$jsonResponseObj"
		
		# Get the session id
		$session = $jsonResponseObj.session
		
		# ================================================================
		# 1.2 AUTHENTICATE THE LOGIN

		# Create MD5 hash-encoded credentials string
		$response_str = $bi_user_name + ":" + $session + ":" + $bi_user_pw
		$response_hash = encryptMD5 $response_str

		# Create, then encode a modified JSON 'login' command
		if ($dev_uuid) {
			$parameters = @{
				cmd = 'login'
				session = $session
				response = $response_hash
				uuid = $dev_uuid
			}
		} else {
			$parameters = @{
				cmd = 'login'
				session = $session
				response = $response_hash
			}
		}
		$json_params = $parameters | ConvertTo-Json

		# For debugging
		#Write-Host "$fn_name() (1.2) authorizing json login... `n$json_params"
		#msg.exe $server_username /TIME:15 "$fn_name() (1.2) authorizing json login ... `n$json_params"
		
		# Execute the JSON request
		$jsonResponseObj = $json_params | Invoke-RestMethod -Uri $uri -Method Post

		# For debugging
		#Write-Host ("$fn_name() (1.2) jsonResponse... `n" + ($jsonResponseObj  | ConvertTo-Json))
		#msg.exe $server_username /TIME:15 "$fn_name() (1.2) jsonResponse... `n$jsonResponseObj"
		
		# Handle failure
		if ($jsonResponseObj.result -eq 'fail') {
			$err_msg = "ERR/unable to authenticate Blue Iris JSON interface`n"
			return $err_msg
		}
		
		# Get the authenticated session id
		$auth_session_id = $jsonResponseObj.session

		# For debugging
		#msg.exe $server_username /TIME:5 "$fn_name() auth_session_id... $auth_session_id"
	
		# Save the authenticated session id
		# NOTE: the session id is always saved, whether used or not
		$auth_session_id | Set-Content "$session_filespec" -NoNewline

	} else {
	
		# Retrieve the saved session id
		# KEY POINT: 
		#   Force typing the output of 'Get-Content' to [String] was essential when running script in
		#   in Powershell ISE. Otherwise the output was similar to an object.
		#   *** But in actuality it is more complicated ***
		#   Per https://stackoverflow.com/a/55881978/13499984
		#      When 'Get-Content' is assigned to a variable, its output is an array of .NET [String] objects, 
		#      each representing a line from the input file. However, Get-Content also decorates these string 
		#      objects with additional properties that provide helpful metadata about where the strings came from.
		$saved_session_id = [String](Get-Content -Path "$session_filespec")

		$auth_session_id = $saved_session_id

		# For debugging
		if (!$logging_out_flg) {
			#Write-Host ("$fn_name() saved_session_id type = " + $saved_session_id.GetType())
			#Write-Host "$fn_name() saved_session_id... $saved_session_id"
			#msg.exe * /TIME:5 "$fn_name() saved_session_id... $saved_session_id"
		}
		
	}

	# ================================================================
	# 1.3 EXECUTE THE REQUESTED JSON COMMAND

	if ($dev_uuid -eq $null) {
		
		# In the event that this JSON call fails with reason 'Invalid session',
		# the script will automatically attempt up to $max_retries
	
		# Modify, then encode the requested JSON login command
		$JsonCmd.session = $auth_session_id
		$json_params = $JsonCmd | ConvertTo-Json

		# Execute the JSON request
		$params = @{
			Uri                = $uri
			#Headers            = @{"Content-Type"  = "application/json"}
			Method             = 'POST'
			#Body               = $JsonCmd | ConvertTo-Json
			Body               = $json_params
			ContentType        = 'application/json'
		}
		
		$json_response_obj = Invoke-RestMethod @params

		# Alternate syntaxes (pick one)
		#$json_response_obj = Invoke-RestMethod -Uri $uri -Method Post -Body ($JsonCmd | ConvertTo-Json) #>
		#$json_response_obj = ($JsonCmd | ConvertTo-Json) | Invoke-RestMethod -Uri $uri -Method Post

		# For debugging
		#if ($JsonCmd.cmd -match 'update') {
			#Write-Host ("$fn_name() (1.3) requested JsonCmd '" + $JsonCmd.cmd + "'")
			#Write-Host ("$fn_name() (1.3) requested JsonCmd '" + $JsonCmd.cmd + "' ...`n" + $json_params)
			#Write-Host ("$fn_name() json_response_obj = `n" + ($json_response_obj | ConvertTo-Json))
		#}

		if ($json_response_obj.result -eq 'fail') {
			
			# For debugging
			#Write-Host ("$fn_name() (1.3) JsonCmd '" + $JsonCmd.cmd + "' response.result = fail")
			#Write-Host ("$fn_name() (1.3) JsonCmd '" + $JsonCmd.cmd + "' response = `n" + ($json_response_obj | ConvertTo-Json))

			# Try to determine reason
			if ($json_response_obj.data) {

				$data_obj = $json_response_obj.data

				# For debugging
				#Write-Host ("$fn_name() JsonCmd '" + $JsonCmd.cmd + "' response.data = " + ($data_obj | ConvertTo-Json))
				
				if ($data_obj.reason) {
					$reason = $data_obj.reason
				
					# For debugging
					#Write-Host ("$fn_name() JsonCmd '" + $JsonCmd.cmd + "' response.data.reason = $reason")

					if ($reason -match 'Invalid session') {
						# Update retry counter
						$retry_cntr++
						
						# RECURSIVE CALL
						#Start-Sleep -Milliseconds 500
						$json_response_obj = `
							BIJsonCall `
							-JsonCmd $JsonCmd `
							-ForceFlg $true `
							-RetryNo $retry_cntr `
							-QuietFlg $QuietFlg `
							-FromConsoleFlg $FromConsoleFlg
						
					} elseif ($reason -match 'Access denied') {
						$err_msg ="ERR/JsonCmd '" + $JsonCmd.cmd + "' requires an Administrator account"
						return $err_msg
						
					} else {
						$err_msg = "ERR/JsonCmd '" + $JsonCmd.cmd + "' failed with reason '" + $reason + "'"
						return $err_msg
					}
					
				} else {
					$err_msg = "ERR/JsonCmd '" + $JsonCmd.cmd + "' failed without a reason"
					return $err_msg
				}
			}
		} 
	}

	# For debugging
	if ($json_response_obj) {
		#Write-Host ("$fn_name() (1.3) jsonResponse type = " + $json_response_obj.GetType())
		#Write-Host ("$fn_name() (1.3) jsonResponse... `n" + ($json_response_obj  | ConvertTo-Json))
		#msg.exe $server_username /TIME:5 "$fn_name() (1.3) jsonResponse... $json_response_obj"
	} 

	# Retain for error testing
	#$json_response_obj = $null
	#$json_response_obj = ''

	if ($json_response_obj -eq $null) {
		$err_msg ="ERR/null response from Blue Iris JSON interface"
		return $err_msg
	} 
	if ($json_response_obj -eq '') {
		$err_msg = "ERR/empty response from Blue Iris JSON interface"
		return $err_msg
	}
	if ($json_response_obj.result -ne $null) {

		# For debugging
		#Write-Host ("$fn_name() (1.3) json_response_obj.result... " + ($json_response_obj.result))
		
		if ($json_response_obj.result -eq 'fail') {
			
			# Show the contents of the $json_response_obj 'data' key

			$json_data_obj = $json_response_obj.data
			
			# For debugging
			#Write-Host ("$fn_name() (1.3) json_response_obj.data... " + ($json_response_obj.data))

			if ($PsVersion -ge 7) {
				# Avoids the following WARNING...
				# "Resulting JSON is truncated as serialization has exceeded the set depth of N"
				$jdo_str = $json_data_obj | ConvertTo-Json -Depth 16
			} else {
				$jdo_str = $json_data_obj | ConvertTo-Json
			}
			
			# 'Minify' the json string.
			if ($jdo_str) {
				if (!($jdo_str -match 'null')) {
					$err_msg = "ERR/JSON response for cmd '" + $JsonCmd.cmd + "' is 'result:fail'"
					return $err_msg
				} else {					
					$err_msg = "ERR/JSON response for cmd '" + $JsonCmd.cmd + "' is null (DB record not found)"
					return $err_msg
				}
			}
		}
	}

	return $json_response_obj

}

function BIJsonCamListParse() {
# PURPOSE:   Get Blue Iris web server JSON 'camlist' command response (for all or subset of BI cameras)
#            Intended to be called by other PowerShell scripts.
# ARGUMENTS: -CamProps = [OPTIONAL] (list) like 'bps,bps2', comma delimited
# RETURNS:  (object) PowerShell hashtable, @('cam':camName,'camProp':value}
#		                                   @{camName:{[camProp1:value1],[camProp2:value2],...},...}"
# REQUIRES:  function BIJsonCall()
#
# Description of JSON cmd 'camlist' from the Blue Iris Help PDF:
# 	"returns a list of cameras on the system ordered by group. Cameras not belonging to any
#	group are shown beneath the "all cameras" group. Disabled cameras are placed at the end of
#	the list. An array of objects is returned in data (note the [] surrounding a JSON array), each
#	describing a camera or a camera group."
#
# Version: 1
# Created by jaydeel at ipcamtalk.com

	# MUST BE THE FIRST LINE IN A SCRIPT
	param(
		[string]$CamProps,  # [OPTIONAL] list
		[boolean]$QuietFlg,
		[boolean]$FromConsoleFlg
	)

	$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

	# For debugging
	#Write-Host "Entered $fn_name() ..."
	#Write-Host "$fn_name() CamProps = $CamProps"

	#Set-StrictMode  -Version 3  # https://tinyurl.com/mrx3w9v4
	#Start-Transcript  # default destination is C:\Users\{user}l\Documents

	# PARSE COMMANDLINE ARGS
	if ($CamProps) {
		$cam_props_ary = $CamProps -Split ','
		#msg.exe * /TIME:5 "camProps... $cam_props_ary"
	} else {
		$cam_props_ary = @()
	}
	#Write-Host "cam_props_ary = $cam_props_ary"

	# PERFORM API CALLS
	$jsonCmd = @{
		cmd = 'camlist'
	}
	# Returns a powershell PSCustomObject
	$json_response_obj = `
		BIJsonCall `
		-JsonCmd $jsonCmd `
		-QuietFlg $QuietFlg `
		-FromConsoleFlg $FromConsoleFlg

	# For debugging
	#Write-Host ("$fn_name() BIJsonCall() json_response_obj type = " + $json_response_obj.GetType())
	#Write-Host ("$fn_name() BIJsonCall() json_response_obj = " + ($json_response_obj | ConvertTo-Json))
	#msg.exe * /TIME:10 "$fn_name() BIJsonCall() json_response... $json_response_obj"

	# Abort if GET failed
	$fail_msg = "ERR/JSON request failed"
	if ($json_response_obj -match 'ERR/') {
		if ($fromConsole) {
			# For debugging
			#Write-Host "$fn_name() BIJsonCall() json_response_obj = $json_response_obj`n"
		} else {
			#msg.exe * /TIME:10 "$fn_name() BIJsonCall() json_response_obj = $json_response_obj"
		}
		return $fail_msg
	}
	if ($json_response_obj.result -eq 'fail') {
		$x  = $json_response_obj.result
		if ($fromConsole) {
			# For debugging
			#Write-Host "$fn_name() BIJsonCall() json_response_obj.result = $x`n"
		} else {
			#msg.exe * /TIME:10 "$fn_name() BIJsonCall() json_response_obj.result = $x"
		}
		return $fail_msg
	}

	# Extract data from the JSON command response
	$data_obj = $json_response_obj.data

	# For error testing
	#$data_obj = $null

	# Extract property value
	# ... Note: null/blank $data_obj.ram observed on occasion
	if ($data_obj) {
		$data_cnt = $data_obj.Count
		
		# For debugging
		#Write-Host ("data_obj cnt = " + $data_cnt)
		#Write-Host ("data_obj[0] = `n" + $data_obj[0])
		#Write-Host ("data_obj[" + ($data_cnt - 1) +"] = `n" + $data_obj[$data_cnt-1])

		$rtn_obj = [ordered]@{}

		for ($i=0; $i -lt $data_obj.length; $i++) { 
			$item_obj = $data_obj[$i]
			$option_disp = $item_obj.optionDisplay;
			$option_val  = $item_obj.optionValue;
			
			# For debugging
			#Write-Host "$i : option_val = $option_val, option_disp = $option_disp"
			#Write-Host ("    option_val starts w/ @ = " + $option_val.StartsWith("@"))
			#Write-Host ("    option_disp starts w/ + = " + $option_disp.StartsWith("+"))
			
			if ($option_val -ne "Index" -and !$option_disp.StartsWith("+") -and !$option_val.StartsWith("@")) {
				$new_obj = [ordered]@{}
				
				if ($cam_props_ary.length -gt 0) {
					
					# GET REQUESTED PROPERTIES
					
					foreach ($cam_prop in $cam_props_ary) {
						#Write-Host ("cam_prop = $cam_prop, val = " + $item_obj.$cam_prop)
						if ($item_obj.$cam_prop -or $item_obj.$cam_prop -eq 0) {
							$x = $item_obj.$cam_prop
							$new_obj.Add($cam_prop, $x)
							#$new_obj.$cam_prop = $x
						} else {
							$new_obj.Add($cam_prop, $null)
							#$new_obj.$cam_prop = $null
						}
					}
				} else {
					
					# GET ALL PROPERTIES
					
					# Get all property names in the JSON response (a PowerShell PSCustomObject)
					# ... NOTE: Extracting 'keys' from the PSCustomObject was bit tricky
					#           REF: https://stackoverflow.com/a/18780001/13499984
					$item_obj_keys = @()
					foreach ($item in $item_obj) {
						$item_obj_keys += $($item | Get-Member -MemberType *Property).Name
					}

					# For debugging
					#Write-Host ("item_obj_keys keys = " + $item_obj_keys)
					
					foreach ($cam_prop in $item_obj_keys) {
						#Write-Host ("cam_prop = $cam_prop, val = " + $item_obj.$cam_prop)
						if ($item_obj.$cam_prop -or $item_obj.$cam_prop -eq 0) {
							$x = $item_obj.$cam_prop
							$new_obj.Add($cam_prop, $x)
						} else {
							$new_obj.Add($cam_prop, $null)
						}
					}
				}
				$rtn_obj[$option_val] = $new_obj
			}
		}

		# @@@ DEV - Retain for testing
		#Write-Host ("rtn_obj cnt = " + $rtn_obj.count)
		#Write-Host "rtn_obj[DW1] = "$rtn_obj.DW1
		#Write-Host "rtn_obj[DW1]['BPS'] = "$rtn_obj['DW1']['BPS']
		#Write-Host "rtn_obj keys = "$rtn_obj.Keys
		
	} else {
		#Write-Host "sorry - JSON query returned null/blank`n"
	}

	return $rtn_obj
}

function Write-ColorOutput($ForegroundColor) {
# Source: https://stackoverflow.com/a/4647985
# Examples:  Write-ColorOutput Red (string)
#            Write-ColorOutput Green (string)
#            string | Write-ColorOutput Yellow

    # save the current color
    $fc = $host.UI.RawUI.ForegroundColor

    # set the new color
    $host.UI.RawUI.ForegroundColor = $ForegroundColor

    # output
    if ($args) {
        Write-Output $args
    }
    else {
        $input | Write-Output
    }

    # restore the original color
    $host.UI.RawUI.ForegroundColor = $fc
}

# FEEDBACK UNEXPECTED COMMANDLINE ARGUMENTS
if (1 -eq 1) {  # Use to collapse this code section while editing
	#Write-Host ("params = " + ($PSBoundParameters | ConvertTo-Json))
	$cmdline_args_passed = [String[]]$PSBoundParameters.Keys
	#Write-Host "Extra Args w/ values = $Args"
	$cmdline_args_bad = ($Args | where{$_ -like "-*"}) | ForEach-Object{$_ -Replace '-', ''}
	if ($cmdline_args_bad) {
		$cmdline_args_bad = [String[]]$cmdline_args_bad
	} else {
		$cmdline_args_bad = @()
	}
	#Write-Host "Cmdline Args Passed = $cmdline_args_passed"
	#Write-Host "Cmdline Args Bad = $cmdline_args_bad"
	if ($cmdline_args_bad.Count -gt 0 -and !$Help -and !$Examples) {
		if ($fromConsole) {
			Write-Host ("HEADS-UP: commandline contains unsupported argument(s): [ " + ($cmdline_args_bad -Join ', ') + " ]") -ForegroundColor Green
			return
		}
		
	}
}

# SHOW USAGE GUIDE
if ($Help) {
	if ($fromConsole) {
		Write-Host "======================================================================================"
		Write-Host "BLUE IRIS CLIPS DISK USAGE UTILITY"
		Write-Host "- Tabulates disk usage in each Blue Iris-managed folder [1]"
		Write-Host "- Shows usage by camera AND file type (bvr, mp4, jpg, dat); saves results to CSV file"
		Write-Host "- Optionally timestamps the CSV files to archive clip usage changes over time"
		Write-Host "--------------------------------------------------------------------------------------"
		Write-Host "Usage: $ScriptName {no args} ... show disk usage tables"
		Write-Host "       $ScriptName -ShowFolders ... review Blue Iris folders to be analyzed"
		Write-Host "Args:"
		Write-Host "   -SortColNo ... sort output tables on specified column number (1-N)"
		Write-Host "   -Ascending ... sort column in ascending order (default is descending)"
		Write-Host "   -DefFnameFormat ... force script to assume the default filename format [2]"
		Write-Host "   -NoTimestamp ... disable time stamping of output CSV files [3]"
		Write-Host "   -NoPruning ... disable automatic pruning of old CSV output files [4]"
		Write-Host "   -ClearArchive ... PERMANENTLY delete all old timestamped CSV files [5]"
		Write-Host "   -ShowFolders ... display all Blue Iris folders & highlight those being analyzed"
		Write-Host "   -Help ... show this usage guide"
		Write-Host "Notes: "
		Write-Host "[1] User-configurable via user-setting-file '`$show_folderN', where N=1 to 16"
		Write-Host "[2] This setting is specified in each BI Camera Settings > Recording tab"
		Write-Host "    (by default, the script auto-determines the default filename format state)"
		Write-Host "[3] Overrides user setting file '`$timestamp_output_csv_files_flg'"
		Write-Host "[4] On-demand equivalent to user-setting-file '`$archive_lifetime_days=0'"
		Write-Host "[5] On-demand equivalent to user-setting-file '`$archive_lifetime_days=-1'"
		Write-Host "    WARNING - the deleted files are not recoverable (not saved to the Recycle Bin)"
		Write-Host "--------------------------------------------------------------------------------------"
		Write-Host ""
		exit
	}
}

$timer_script = [Diagnostics.Stopwatch]::StartNew()

# ==============================================================
# IMPORT USER SETTINGS...
# Note: The 'user settings' script is expected to be in same folder as this script
$user_settings_filename = $ScriptName.Replace(".ps1", "_user_settings.ps1")
try {
	. $ScriptPath\$user_settings_filename
	
	# For debugging
	#Write-Host "Loaded user settings from '$user_settings_filename'"
}
catch {
	$err_msg = "ERR/failed to load user settings file '$user_settings_filename'`nTip: try executing it directly to confirm it exists and reveal syntax errors"
	return $err_msg
}
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# PARSE USER SETTINGS
if (1 -eq 1) {  # Use to collapse this code section while editing

	$variable_test = Get-Variable -Name app_path -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$err_msg = "ERR/user setting '`$app_path' is missing"
		return $err_msg
	}
	if ($app_path.StartsWith('YOUR_')) {  
		$err_msg = "ERR/user-setting file setting 'app_path' is not set"
		return $err_msg
	}

	$variable_test = Get-Variable -Name experienced_user_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$experienced_user_flg = $false
	}

	$variable_test = Get-Variable -Name default_filename_format_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$override_default_filename_format_check_flg = $false
	} else {
		$override_default_filename_format_check_flg = $true
	}
	
	if (!$override_default_filename_format_check_flg) {

		$variable_test = Get-Variable -Name bi_webserver -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$err_msg = "ERR/user setting '`$bi_webserver' is missing"
			return $err_msg
		}
		if ($bi_webserver.StartsWith('YOUR_')) {  
			$err_msg = "ERR/user-setting file setting 'bi_webserver' is not set"
			return $err_msg
		}
		$variable_test = Get-Variable -Name bi_user_name -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$err_msg = "ERR/user setting '`$bi_user_name' is missing"
			return $err_msg
		}
		if ($bi_user_name.StartsWith('YOUR_')) {
			$err_msg = "ERR/user-setting file setting 'bi_user_name' is not set"
			return $err_msg
		}
		$variable_test = Get-Variable -Name bi_user_pw -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$err_msg = "ERR/user setting '`$bi_user_pw' is missing"
			return $err_msg
		}
		if ($bi_user_pw.StartsWith('YOUR_')) {
			$err_msg = "ERR/user-setting file setting 'bi_user_pw' is not set"
			return $err_msg
		}
	}

	$variable_test = Get-Variable -Name timestamp_output_csv_files_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$timestamp_output_csv_files_flg = $true
	}

	$variable_test = Get-Variable -Name archive_lifetime_days -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$archive_lifetime_days = 365
	}
	
	$show_folders = @()
	for ($i=1; $i -le 16; $i++) { 

		# Get user-setting variable $show_folderN
		$show_flg = [boolean](Get-Variable -Name "show_folder$i" -ValueOnly -ErrorVariable errs -ErrorAction SilentlyContinue)
		
		# For debugging
		#Write-Host "show_flg = '$show_flg'"
		#Write-Host ("errs cnt = " + $errs.Count)
		#Write-Host "errs = $errs"

		if ($errs.Count -eq 0) {
			# Variable found; so set to user's value
			$show_folders += $show_flg
		} else {
			# Variable not found; so set default values
			if ($i -lt 4) {
				$show_folders += $true
			} else {
				$show_folders += $false
			}
		}
	}
	# For debugging
	#Write-Host ("show_folders = " + ($show_folders | ConvertTo-Json))

	$override_folders = @()
	for ($i=1; $i -le 16; $i++) { 

		# Get user-setting variable $override_folderN
		$override_folder = (Get-Variable -Name "override_folder$i" -ValueOnly -ErrorVariable errs -ErrorAction SilentlyContinue)
		
		# For debugging
		#Write-Host "show_flg = '$show_flg'"
		#Write-Host ("errs cnt = " + $errs.Count)
		#Write-Host "errs = $errs"

		if ($errs.Count -eq 0) {
			# Variable found; so set to user's folder, if it exists
			if (Test-Path -LiteralPath $override_folder) {
				$override_folders += $override_folder
				Write-Warning "per USER SETTING, using override folder #$i '$override_folder' instead of default"
			} else {
				$override_folders += ""
				Write-Warning "ignoring USER SETTING, override folder #$i ($override_folder); folder does not exist"
			}
		} else {
			# Variable not found; so set default values
			$override_folders += ""
		}
	}
	# For debugging
	#Write-Host ("override_folders = " + ($override_folders | ConvertTo-Json))

}

# CHECK IF SCRIPT'S PATH IS IN ENVIRONMENT 'SYSTEM PATH' VARIABLE
if (!$experienced_user_flg) {
	if ($FromConsole) {

		$EnvPath = [Environment]::GetEnvironmentVariable('Path', 'Machine')

		# For debugging
		# Write-Host "ScriptPath = $ScriptPath"
		# Write-Host ("EnvPath = " + (($EnvPath -Split ';') | ConvertTo-Json))

		if ($null -ne $EnvPath) {
			if (!($EnvPath -split ';' -contains $ScriptPath)) {

				#return "ERR/this script's path '$ScriptPath' does not exist in the Environment System Path variable"
				Write-Host "HEADS-UP: this script's path '$ScriptPath' does not exist in the Environment System Path variable" -ForegroundColor Green
				
				# Attempt repair
				# ABORTED: when I tried this it added MANY other/duplicate entries
				#[Environment]::SetEnvironmentVariable("Path", $EnvPath + ";$ScriptPath", "Machine")
				#Write-Warning "Added this script's path '$ScriptPath' to the Environment Path"
			}
		} else {
			#Write-Warning "Unable to inspect Environment variable; be sure script's path is in the Environment Path"
			Write-Host "HEADS-UP: Unable to inspect Environment variable; be sure script's path is in the Environment Path" -ForegroundColor Green
		}

	}
}

# PARSE COMMANDLINE ARGS
if (1 -eq 1) {  # Use to collapse this code section while editing

	if ($SortColNo) {
		# Test if argument contains non-numeric characters
		$num_str = ($SortColNo -replace "[^0-9]" , '').ToString()
		if ($num_str.Length -eq 0) {
			$err_msg = "ERR/argument -SortColNo must be an integer"
			return $err_msg
		}
	}

	if ($PSBoundParameters.Keys -contains 'NoTimestamp') {
		$timestamp_output_csv_files_flg = $false
	}

	if ($PSBoundParameters.Keys -contains 'NoPruning') {
		$archive_lifetime_days = 0
	}

	if ($PSBoundParameters.Keys -contains 'ClearArchive') {
		$archive_lifetime_days = -1
	}

	if ($PSBoundParameters.Keys -contains 'DefFnameFormat') {
		$default_filename_format_flg = $True
		$override_default_filename_format_check_flg = $true
		Write-Warning "per Arg -DefFnameFormat, script assumes clips are using the DEFAULT filename format "
	} else {
		if ($override_default_filename_format_check_flg) {
			if ($default_filename_format_flg) {
				Write-Warning "per USER SETTING, script assumes clips are using the DEFAULT filename format"
			} else {
				Write-Warning "per USER SETTING, script assumes clips are using a NON-DEFAULT filename format"
			}
		} else {
			$default_filename_format_flg = $true
		}
	}
}

# For testing
#$default_filename_format_flg = $true;

# GET BLUE IRIS STORAGE FOLDERS
# ... Initialize vars
$alerts_folder_path = ""
$bi_folder_paths = @()  # Initialize accumulator (all Blue Iris Storage folder paths)
$sources = @()  # Accumulator for folders to be processed/displayed
# ... Registry location
$key_path = 'HKLM:\SOFTWARE\Perspective Software\Blue Iris\clips\folders'
$key_name = 'path'
# ... Parse all 16 built-in folders
for ($i = 0; $i -le 15; $i++) {

	# GET REGISTRY VALUE
	$key_path_N = $key_path + $PathSep + $i
	$key_val = `
		RegGetValue `
		-KeyPath $key_path_N `
		-KeyName $key_name
		
	# For debugging
	#Write-Host "$key_path_N = $key_val"
				
	# Trap errors
	if ($null -ne $key_val) {
		if ($key_val.GetType() -match 'string') {
			if ($key_val.StartsWith('ERR/')) {
				#$err_msg = $key_val
				#return $err_msg
				$key_val = ''
			}
		}
	}
 
	# Load accumulator
	$bi_folder_paths += $key_val
 
	# Proceed only if user opted in to show this folder
	$show_flg = $show_folders[$i]
	if ($show_flg -or $i -eq 2) {

		# Proceed only if 'path' found in Registry
		if ($null -ne $key_val) {
			
			# Trap errors
			if ($key_val.GetType() -match 'string') {
				if ($key_val.StartsWith('ERR/')) {
					$err_msg = $key_val
					return $err_msg
				}
			}
			
			# Load accumulator
			if ($show_flg) {
				$sources += $key_val
			}
			
			# ID the (special) Alerts folder
			if ($i -eq 2) {
				$alerts_folder_path = $key_val
			}
		} else {
			Write-Warning ("ignoring folder #$i; Blue Iris folder '" + $bi_folder_paths[$i] + "' does not have a defined path")
		}
	}

}

# For debugging
#Write-Host ("folders from Registry = " + ($bi_folder_paths | ConvertTo-Json))
#Write-Host ("folders to show = " + ($sources | ConvertTo-Json))
#Write-Host "alerts_folder_path = $alerts_folder_path"

# REPLACE USER-SPECIFIED OVERRIDE FOLDERS
# NOTE: override folders already verified
for ($i = 0; $i -le 15; $i++) {

	$show_flg = $show_folders[$i]
	$override_folder = $override_folders[$i]

	# Proceed only if user opted in to show this folder
	if ($show_flg -and $override_folder.Length -gt 0) {
		$sources[$i] = $override_folder
	}

}

# For debugging
#Write-Host ("folders w/ overrides = " + ($sources | ConvertTo-Json))

# DISPLAY ALL BLUE IRIS STORAGE FOLDERS - AND HIGHLIGHT THOSE BEING ANALYZED
if ($ShowFolders) {
	if ($fromConsole) {
		if ($host.name -match 'ISE') {			
			Write-Host "`nBlue Iris Storage folders (from the registry):`n(only the *** folders will be analyzed by this script)"
		} else {
			Write-Host "`nBlue Iris Storage folders (from the registry):`n(only the green folders will be analyzed by this script)"
		}
		Write-Host "--------------------------------------------------------"
		for ($i=1; $i -le 16; $i++) { 
			$pfx = ("`$folder" + $i + " (" + $bi_folders[$i-1] + ")").PadRight(17," ")
			if ($show_folders[$i-1]) {
				if ($host.name -match 'ISE') {			
					$str = $pfx + " = '" + $bi_folder_paths[$i-1] + "'"
					if ($override_folders[$i-1]) {
						$str += " NOTE: user override as '" + $sources[$i-1] + "'"
					}
					$str += " ***"
					Write-Host $str
				} else {
					$str = $pfx + " = '" + $bi_folder_paths[$i-1] + "'"
					if ($override_folders[$i-1]) {
						$str += " NOTE: user override as '" + $sources[$i-1] + "'"
					}
					Write-ColorOutput Green $str
				} 
			} else {
				$str = $pfx + " = '" + $bi_folder_paths[$i-1] + "'"
				Write-Host $str
			}
		}
	}
	Write-Host ""
	exit
}

# DETERMINE IF ANY CAMERAS ARE NOT USING THE DEFAULT FILE NAMING FORMAT (PATTERN) ON THE RECORD TAB

# ... By default, the script assumes a QUALIFIED non-default filename format is being used by Blue Iris
#     if the following conditions are true FOR ANY CAMERA:
#     1. The camera is ENABLED
#     2. Either 'Video' of 'JPGs' recording is enabled (Camera settings > Record tab)
#     3. The 'Use default' folder setting is NOT CHECKED
#     4. The custom filename format INCLUDES SUB-FOLDERS ... e.g. '&CAM\%d-%m-%y\&CAM.%Y%m%d_%H%M%S'
#     5. The final element of the custom filename format starts or ends with '&CAM'
#     6. the segment at the opposite end of the filename must begin with a BI datetime macro ('%' character)
#     NOTE: per the Blue Iris Help PDF, 
#           A. The actual filename portion of the path should always either begin or end with 
#              the camera name, either &CAM.xxxx or xxxx.&CAM
#           B. the short name must not begin with a number, while the segment at the opposite end 
#              of the filename must begin with a number (typically part of the date or time)
# Requested by joshwa https://ipcamtalk.com/threads/blue-iris-clips-and-files-disk-usage-utility.67655/post-803999
#    for format &CAM.\%d-%m-%y.\&CAM.%Y%m%d_%H%M%S
#

# ... skip registry interrogation if unneeded
if (!$override_default_filename_format_check_flg) {

	# Read each camera's settings
	$cliplist_hashtable = `
		BIJsonCamListParse `
		-CamProps 'optionDisplay,isEnabled' `
		-QuietFlg $false `
		-FromConsoleFlg $fromConsole
			
	# For debugging
	#Write-Host ("hashtable keys = " + $cliplist_hashtable.Keys)
	#Write-Host ("hashtable = " + ($cliplist_hashtable | ConvertTo-Json))

	$cams_ary = $cliplist_hashtable.Keys
	$cam_lognnames_ary = @()
	foreach ($item in $cams_ary) {
		if ($cliplist_hashtable[$item]['isEnabled']) {
			$cam_lognnames_ary += $cliplist_hashtable[$item]['optionDisplay']
		}
	}
			
	# For debugging
	#Write-Host ("cam_lognnames_ary = " + ($cam_lognnames_ary | ConvertTo-Json))

	# Start timer
	$timer_reg = [Diagnostics.Stopwatch]::StartNew()
	
	$i = 0
	foreach ($item in $cam_lognnames_ary) {  # All enabled cameras
		
		for ($j =1; $j -le 7; $j++) {  # Parse each profile

			# GET REGISTRY VALUES
			$key_path = "HKLM:\SOFTWARE\Perspective Software\Blue Iris\Cameras\$item\Clips\$j"

			$key_name_1 = "continuous"
			$key_val_1 = `
				RegGetValue `
				-KeyPath $key_path `
				-KeyName $key_name_1

			$key_name_1A = "defformat"
			$key_val_1A = `
				RegGetValue `
				-KeyPath $key_path `
				-KeyName $key_name_1A

			$key_name_1B = "filename"
			$key_val_1B = `
				RegGetValue `
				-KeyPath $key_path `
				-KeyName $key_name_1B

			$key_name_2 = "snapimages"
			$key_val_2 = `
				RegGetValue `
				-KeyPath $key_path `
				-KeyName $key_name_2

			$key_name_2A = "defformat2"
			$key_val_2A = `
				RegGetValue `
				-KeyPath $key_path `
				-KeyName $key_name_2A

			$key_name_2B = "filename2"
			$key_val_2B = `
				RegGetValue `
				-KeyPath $key_path `
				-KeyName $key_name_2B
				
 			# For debugging
			# Write-Host ($key_path + "\" + $key_name_1 + " = " + $key_val_1)
			# Write-Host ($key_path + "\" + $key_name_1A + " = " + $key_val_1A)
			# Write-Host ($key_path + "\" + $key_name_1B + " = " + $key_val_1B)
			# Write-Host ($key_path + "\" + $key_name_2 + " = " + $key_val_2)
			# Write-Host ($key_path + "\" + $key_name_2A + " = " + $key_val_2A)
			# Write-Host ($key_path + "\" + $key_name_2B + " = " + $key_val_2B)
			# Write-Host ("backslash count = " + ([Regex]::Matches($key_val_2B, "\\").Count))
			
			$test1 = ($null -ne $key_val_1 -and $null -ne $key_val_1A -and $null -ne $key_val_1B)
			$test2 = ($null -ne $key_val_2 -and $null -ne $key_val_2A -and $null -ne $key_val_2B)
			if ($test1 -and $test2) {
				
				# Trap errors
				if ($key_val_1.GetType() -match 'string') {
					if ($key_val_1.StartsWith('ERR/')) {
						$err_msg = $key_val_1
						return $err_msg
					}
				}
				if ($key_val_1A.GetType() -match 'string') {
					if ($key_val_1A.StartsWith('ERR/')) {
						$err_msg = $key_val_1A
						return $err_msg
					}
				}
				if ($key_val_1B.GetType() -match 'string') {
					if ($key_val_1B.StartsWith('ERR/')) {
						$err_msg = $key_val_1B
						return $err_msg
					}
				}
				if ($key_val_2.GetType() -match 'string') {
					if ($key_val_2.StartsWith('ERR/')) {
						$err_msg = $key_val_2
						return $err_msg
					}
				}
				if ($key_val_2A.GetType() -match 'string') {
					if ($key_val_2A.StartsWith('ERR/')) {
						$err_msg = $key_val_2A
						return $err_msg
					}
				}
				if ($key_val_2B.GetType() -match 'string') {
					if ($key_val_2B.StartsWith('ERR/')) {
						$err_msg = $key_val_2B
						return $err_msg
					}
				}

				# Test if: 1) recording video, 2) not using default format, & 3) custom format contains a subfolder
				# ... NOTE: key_val_1 = 0=When Triggered; 1=Continuous; 2=Off, 3=Periodic each; 4=Triggered+Periodic: each; 
				#                       5=Continuous+Alerts; 6=When alerted; 7=Continuous+Triggered 
				$test1A = (($key_val_1 -ne 2) -and ($key_val_1A -ne 1) -and ([Regex]::Matches($key_val_1B, "\\").Count -gt 0))
				
				# Test if the actual filename portion of the path either begins or ends with the camera name, either &CAM.xxxx or xxxx.&CAM
				$arr1 = $key_val_1B -Split '\\'
				$test1B = ($arr1[-1].StartsWith('&CAM') -or $arr1[-1].EndsWith('&CAM'))

				# Test if the segment at the opposite end of the filename begins with a number (typically part of the date or time)
				# ... in this the number will be  Blue Iris datetime macro, like %Y, $m. %d, %H, %M, %S
				$test1C = $false
				if ($arr1[-1].StartsWith('&CAM')) {
					$last_part = $arr1[-1].Replace('&CAM','').Trim('.')
					$arr2 = $last_part -Split '' -ne ''   # Remove terminal blank elements
					if ($arr2[0] -match "%") {
						$test1C = $true
					}					
				}
				if ($arr1[-1].EndsWith('&CAM')) {
					$arr2 = $arr1[-1] -Split '' -ne ''   # Remove terminal blank elements
					if ($arr2[0] -match "%") {
						$test1C = $true
					}
				}

				# For debugging
				# if ($arr1.Count -gt 1) {
					# Write-Host ("arr1 = " + ($arr1 | ConvertTo-Json))
					# Write-Host ("arr1[-1] = '" + $arr1[-1] + "'")
					# Write-Host ("last_part = '" + $last_part + "'")
					# Write-Host ("arr2 = " + ($arr2 | ConvertTo-Json))
					# Write-Host ("arr2[0] = '" + $arr2[0] + "'")
					# Write-Host "key_val_1 = $key_val_1, key_val_1A = $key_val_1A, key_val_1B = $key_val_1B "
					# Write-Host ("key_val_1 -ne 2 = " + ($key_val_1 -gt 0) + ", key_val_1A -ne 1 = " + ($key_val_1A -ne 1) + ", key_val_1B '\' cnt -gt 0 = " + ([Regex]::Matches($key_val_1B, "\\").Count -gt 0))
					# Write-Host "test1A = $test1A, test1B = $test1B, test1C = $test1C"
				# }			

				# Test if: 1) recording JPGs, 2) not using default format, & 3) custom format contains a subfolder
				$test2A = (($key_val_2 -gt 0) -and ($key_val_2A -ne 1) -and ([Regex]::Matches($key_val_2B, "\\").Count -gt 0))
				
				# Test if the actual filename portion of the path either begins or ends with the camera name, either &CAM.xxxx or xxxx.&CAM
				$arr1 = $key_val_2B -Split '\\'
				$test2B = ($arr1[-1].StartsWith('&CAM') -or $arr1[-1].EndsWith('&CAM'))

				# Test if the segment at the opposite end of the filename begins with a number (typically part of the date or time)
				# ... in this the number will be  Blue Iris datetime macro, like %Y, $m. %d, %H, %M, %S
				$test2C = $false
				if ($arr1[-1].StartsWith('&CAM')) {
					$last_part = $arr1[-1].Replace('&CAM','').Trim('.')
					$arr2 = $last_part -Split '' -ne ''   # Remove terminal blank elements
					#if ($arr2[0] -match "^\d+$") {
					if ($arr2[0] -match "%") {
						$test2C = $true
					}					
				}
				if ($arr1[-1].EndsWith('&CAM')) {
					$arr2 = $arr1[-1] -Split '' -ne ''   # Remove terminal blank elements
					#if ($arr2[0] -match "^\d+$") {
					if ($arr2[0] -match "%") {
						$test2C = $true
					}
				}

				# For debugging
				# if ($arr1.Count -gt 1) {
					# Write-Host ("arr1 = " + ($arr1 | ConvertTo-Json))
					# Write-Host ("arr1[-1] = '" + $arr1[-1] + "'")
					# Write-Host ("last_part = '" + $last_part + "'")
					# Write-Host ("arr2 = " + ($arr2 | ConvertTo-Json))
					# Write-Host ("arr2[0] = '" + $arr2[0] + "'")
					# Write-Host "key_val_2 = $key_val_2, key_val_2A = $key_val_2A, key_val_2B = $key_val_2B "
					# Write-Host ("key_val_2 -gt 0 = " + ($key_val_2 -gt 0) + ", key_val_2A -ne 1 = " + ($key_val_2A -ne 1) + ", key_val_2B '\' cnt -gt 0 = " + ([Regex]::Matches($key_val_2B, "\\").Count -gt 0))
					# Write-Host "test2A = $test2A, test2B = $test2B, test2C = $test2C"
				# }			
				
				# For debugging
				# if ($key_val_2A -ne 1 -or $key_val_2A -ne 1) {
					# Write-Host "$item... key 'continuous' = $key_val_1"
					# Write-Host "$item... key 'defformat' = $key_val_1A"
					# Write-Host ("$item... key 'format' = $key_val_1B ; has '\' count > 0 is: " + ([Regex]::Matches($key_val_1B, "\\").Count -gt 0))
					# Write-Host "$item... key 'snapimages' = $key_val_2"
					# Write-Host "$item... key 'defformat2' = $key_val_2A"
					# Write-Host ("$item... key 'format2' = $key_val_2B ; has '\' count > 0 is: " + ([Regex]::Matches($key_val_2B, "\\").Count -gt 0))
					# Write-Host "$item... test1A = $test1A, test2A = $test2A"
				# }
				
				# Check of all requirements met for the Camera settings > Recording tab custom filename format
				if (($test1A -and $test1B -and $test1C) -or ($test2A -and $test2B -and $test2C)) {
					
					# Custom format is qualified
					$default_filename_format_flg = $false

					$str = ("NOTE: At least one camera (" + $cams_ary[$i] + ") is using a qualified non-default clip default filename format")
					Write-ColorOutput yellow ($str)

					break
				}
			}
			
			if (!$default_filename_format_flg) {
				break
			}
		}
		$i++
	}
	$timer_reg.stop()
	
	# For debugging
	# exit

	if ($DevMode) {
		if ($fromConsole) {
			Write-Host ("(dev mode) registry read time = " + ($timer_reg.elapsed.totalmilliseconds).ToString("0.00") + " msec") -ForegroundColor Cyan
		}
	}

} else {
	#Write-Warning "user setting is overriding automatic determination if using non-default filename pattern (Camera settings > Record tab)"
}

# For debugging
#Write-Host ("default_filename_format_flg = " + $default_filename_format_flg)

if ($fromConsole) {
	if (!$DefFnameFormat -and !$override_default_filename_format_check_flg) {
		if ($default_filename_format_flg) {
			$not_str = ''
			Write-Host "Based on camera Record tab settings (in registry), Blue Iris is using the DEFAULT filename format"
		} else {
			Write-Host "HEADS-UP: Based on camera Record tab settings (in registry), Blue Iris is NOT using the DEFAULT filename format" -ForegroundColor Green
		}
	}
}

# CREATE DIRECTORY FOR OUTPUT (if necessary).
If (!(Test-Path -PathType container $app_path)) {
	New-Item -ItemType Directory -Path $app_path | Out-Null
}

if ($fromConsole) {
	Write-Host "`nReading directories and binning file info..."
}

# Show file type tables for each directory.
$i = 0
foreach ($source in $sources) {
	
	# For debugging
	#Write-Host "source = '$source'"

	# Get info arrays for Blue Iris clip type files in directory (*.bvr, *.mp4, *.jpg, *.dat).
	$get_time_ms, $files_ary, $subdirs_ary, $sizes_ary, $dates_ary, $tstamps_ary = `
		get_files_recursive `
		-Path $source

	# For debugging
	<# if ($source.indexOf('Aux1') -gt -1) {
		Write-Host ("subdirs_ary = " + ($subdirs_ary | ConvertTo-Json))
	} #>

	# Initialize list of unique camera files in directory.
	$cams_list = New-Object System.Collections.Generic.List[System.Object]
	
	# Initialize hashtable collectors to be filled by loop below.
	# ... specific to Alerts and non-Alerts folders
	if ($source -eq $alerts_folder_path) { 

		# The Blue Iris Alerts directory contains different file types than the clips folders.
		$dat_files_tbl    = @{}
		$dat_folders_tbl  = @{}
		$dat_sizes_tbl    = @{}
		$dat_sizes_ro_tbl = @{}
		$dat_dates_tbl    = @{}
		$dat_tstamps_tbl  = @{}
		$dat_days_tbl     = @{}

		# Initialize arrays for all files in each table (i.e., bvr table, mp4 table, jpeg table, data table). 
		$dat_sizes_ary    = @()
		$dat_sizes_ro_ary = @()
		$dat_dates_tbl    = @{}
		$dat_tstamps_ary  = @()
		$dat_days_ary     = @()

	} else {

		$bvr_files_tbl    = @{}
		$bvr_folder_tbl   = @{}
		$bvr_sizes_tbl    = @{}
		$bvr_sizes_ro_tbl = @{}
		$bvr_dates_tbl    = @{}
		$bvr_tstamps_tbl  = @{}
		$bvr_days_tbl     = @{}

		# Initialize arrays for all files in each table (i.e., bvr table, mp4 table, jpeg table, data table). 
		$bvr_sizes_ary    = @()
		$bvr_sizes_ro_ary = @()
		$bvr_dates_tbl    = @{}
		$bvr_tstamps_ary  = @()
		$bvr_days_ary     = @()
	}
		
	# ... common to all folders
	if (1 -eq 1) {  # Use to collapse this code section while editing
	
		$mp4_files_tbl    = @{}
		$mp4_folder_tbl   = @{}
		$mp4_sizes_tbl    = @{}
		$mp4_sizes_ro_tbl = @{}
		$mp4_dates_tbl    = @{}	
		$mp4_tstamps_tbl  = @{}	
		$mp4_days_tbl     = @{}
		$jpg_files_tbl    = @{}
		$jpg_folders_tbl  = @{}
		$jpg_sizes_tbl    = @{}
		$jpg_sizes_ro_tbl = @{}
		$jpg_dates_tbl    = @{}
		$jpg_tstamps_tbl  = @{}
		$jpg_days_tbl     = @{}
		
		# Initialize arrays for all files in each table (i.e., bvr table, mp4 table, jpeg table, data table). 
		$mp4_sizes_ary    = @()
		$mp4_sizes_ro_ary = @()
		$mp4_dates_tbl    = @{}
		$mp4_tstamps_ary  = @()
		$mp4_days_ary     = @()
		$jpg_sizes_ary    = @()
		$jpg_sizes_ro_ary = @()
		$jpg_dates_tbl    = @{}
		$jpg_tstamps_ary  = @()
		$jpg_days_ary     = @()
	}

	# Fill hashtables with file info BY CAMERA (filename, filesize, createdate, timestamp).
	$fkey = 0
	foreach ($file in $files_ary) {  # $file is like 'FD.20231018_050000.bvr'
	
		# Make the full filespec for $file
		$file_path = $source + $PathSep + $file
		if ($default_filename_format_flg) {
			if ($subdirs_ary[$fkey].Length -ne 0) {
				$file_path = $source + $subdirs_ary[$fkey] + $PathSep + $file
			}
		}

		# Get the camera shortname
		$arr = $file.Split(".")
		$cam = $arr[0]

		# Make hashtable key.
		$fcam = $cam  # Default
		if ($default_filename_format_flg) {
			if ($subdirs_ary[$fkey].Length -ne 0) {
				$fcam = $subdirs_ary[$fkey] + $PathSep + $cam
			}
		}

		# Initialize hashtable collectors.
		# ... specific to Alerts and non-Alerts folders
		if ($source -eq $alerts_folder_path) {
			
			if ($dat_files_tbl.Keys -notcontains $fcam) {
				$dat_files_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
			if ($dat_sizes_tbl.Keys -notcontains $fcam) {
				$dat_sizes_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
			if ($dat_sizes_ro_tbl.Keys -notcontains $fcam) {
				$dat_sizes_ro_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
			if ($dat_dates_tbl.Keys -notcontains $fcam) {
				$dat_dates_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
			if ($dat_tstamps_tbl.Keys -notcontains $fcam) {
				$dat_tstamps_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
			if ($dat_days_tbl.Keys -notcontains $fcam) {
				$dat_days_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
			
		} else {
			
			if ($bvr_files_tbl.Keys -notcontains $fcam) {
				$bvr_files_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
			if ($bvr_sizes_tbl.Keys -notcontains $fcam) {
				$bvr_sizes_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
			if ($bvr_sizes_ro_tbl.Keys -notcontains $fcam) {
				$bvr_sizes_ro_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
			if ($bvr_dates_tbl.Keys -notcontains $fcam) {
				$bvr_dates_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
			if ($bvr_tstamps_tbl.Keys -notcontains $fcam) {
				$bvr_tstamps_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
			if ($bvr_days_tbl.Keys -notcontains $fcam) {
				$bvr_days_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
		}
		
		# ... common to all folders
		if (1 -eq 1) {  # Use to collapse this code section while editing
		
			if ($jpg_files_tbl.Keys -notcontains $fcam) {
				$jpg_files_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
			if ($jpg_sizes_tbl.Keys -notcontains $fcam) {
				$jpg_sizes_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
			if ($jpg_sizes_ro_tbl.Keys -notcontains $fcam) {
				$jpg_sizes_ro_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
			if ($jpg_dates_tbl.Keys -notcontains $fcam) {
				$jpg_dates_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
			if ($jpg_tstamps_tbl.Keys -notcontains $fcam) {
				$jpg_tstamps_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
			if ($jpg_days_tbl.Keys -notcontains $fcam) {
				$jpg_days_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
				
			if ($mp4_files_tbl.Keys -notcontains $fcam) {
				$mp4_files_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
			if ($mp4_sizes_tbl.Keys -notcontains $fcam) {
				$mp4_sizes_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
			if ($mp4_sizes_ro_tbl.Keys -notcontains $fcam) {
				$mp4_sizes_ro_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
			if ($mp4_dates_tbl.Keys -notcontains $fcam) {
				$mp4_dates_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
			if ($mp4_tstamps_tbl.Keys -notcontains $fcam) {
				$mp4_tstamps_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
			if ($mp4_days_tbl.Keys -notcontains $fcam) {
				$mp4_days_tbl[$fcam] = New-Object System.Collections.Generic.List[System.Object]
			}
		}
		
		# Load the file info collectors.
		# ... specific to Alerts and non-Alerts folders
		if ($source -eq $alerts_folder_path) {
			
			if ($file.EndsWith(".dat")) {
				$dat_files_tbl[$fcam].Add($files_ary[$fkey])
				$dat_sizes_tbl[$fcam].Add($sizes_ary[$fkey])
				$dat_dates_tbl[$fcam].Add($dates_ary[$fkey])
				$dat_tstamps_tbl[$fcam].Add($tstamps_ary[$fkey])
				$dat_sizes_ary   += $sizes_ary[$fkey]
				$dat_dates_ary   += $dates_ary[$fkey]
				$dat_tstamps_ary += $tstamps_ary[$fkey]
				$dat_days_ary    += $dates_ary[$fkey]
				$file_obj = Get-Item -LiteralPath $file_path
				if ($file_obj.IsReadOnly) {
					$dat_sizes_ro_ary += $sizes_ary[$fkey]
				}
			}
			
		} else {
			
			if ($file.EndsWith(".bvr")) {
				$bvr_files_tbl[$fcam].Add($files_ary[$fkey])
				$bvr_sizes_tbl[$fcam].Add($sizes_ary[$fkey])
				$bvr_dates_tbl[$fcam].Add($dates_ary[$fkey])
				$bvr_tstamps_tbl[$fcam].Add($tstamps_ary[$fkey])
				$bvr_sizes_ary   += $sizes_ary[$fkey]
				$bvr_dates_ary   += $dates_ary[$fkey]
				$bvr_tstamps_ary += $tstamps_ary[$fkey]
				$bvr_days_ary    += $dates_ary[$fkey]
				$file_obj = Get-Item -LiteralPath $file_path
				if ($file_obj.IsReadOnly) {
					$bvr_sizes_ro_ary += $sizes_ary[$fkey]
				}
			}
		}

		# ... common to all folders
		if (1 -eq 1) {  # Use to collapse this code section while editing
		
			if ($file.EndsWith(".jpg")) {
				$jpg_files_tbl[$fcam].Add($files_ary[$fkey])
				$jpg_sizes_tbl[$fcam].Add($sizes_ary[$fkey])
				$jpg_dates_tbl[$fcam].Add($dates_ary[$fkey])
				$jpg_tstamps_tbl[$fcam].Add($tstamps_ary[$fkey])
				$jpg_sizes_ary   += $sizes_ary[$fkey]
				$jpg_dates_ary   += $dates_ary[$fkey]
				$jpg_tstamps_ary += $tstamps_ary[$fkey]
				$jpg_days_ary    += $dates_ary[$fkey]
				$file_obj = Get-Item -LiteralPath $file_path
				if ($file_obj.IsReadOnly) {
					$jpg_sizes_ro_ary += $sizes_ary[$fkey]
				}
			}
			if ($file.EndsWith(".mp4")) {
				$mp4_files_tbl[$fcam].Add($files_ary[$fkey])
				$mp4_sizes_tbl[$fcam].Add($sizes_ary[$fkey])
				$mp4_dates_tbl[$fcam].Add($dates_ary[$fkey])
				$mp4_tstamps_tbl[$fcam].Add($tstamps_ary[$fkey])
				$mp4_sizes_ary   += $sizes_ary[$fkey]
				$mp4_dates_ary   += $dates_ary[$fkey]
				$mp4_tstamps_ary += $tstamps_ary[$fkey]
				$mp4_days_ary    += $dates_ary[$fkey]
				$file_obj = Get-Item -LiteralPath $file_path
				if ($file_obj.IsReadOnly) {
					$mp4_sizes_ro_ary += $sizes_ary[$fkey]
				}
			}
		}

		$fkey++
	}

	# Get total files counts for folder (used for READ-ONLY stats)
	# ... specific to Alerts and non-Alerts folders
	if ($source -eq $alerts_folder_path) {
		$dat_counts_all_total = $dat_sizes_ary.Length
		$dat_counts_ro_total  = $dat_sizes_ro_ary.Length
	} else {
		$bvr_counts_all_total = $bvr_sizes_ary.Length
		$bvr_counts_ro_total  = $bvr_sizes_ro_ary.Length
	}
	# ... common to all folders
	$jpg_counts_all_total = $jpg_sizes_ary.Length
	$jpg_counts_ro_total  = $jpg_sizes_ro_ary.Length
	$mp4_counts_all_total = $mp4_sizes_ary.Length
	$mp4_counts_ro_total  = $mp4_sizes_ro_ary.Length

	# Get total files sizes for folder (used for READ-ONLY stats)
	# ... specific to Alerts and non-Alerts folders
	if ($source -eq $alerts_folder_path) {
		$dat_sizes_all_total = ($dat_sizes_ary | Measure-Object -sum).sum  # bytes
		$dat_sizes_ro_total  = ($dat_sizes_ro_ary | Measure-Object -sum).sum  # bytes
	
		# For debugging
		# Write-Host "dat_counts_ro_total = $dat_counts_ro_total"
		# Write-Host "dat_sizes_ro_total = $dat_sizes_ro_total"
		
	} else {
		$bvr_sizes_all_total = ($bvr_sizes_ary | Measure-Object -sum).sum  # bytes
		$bvr_sizes_ro_total  = ($bvr_sizes_ro_ary | Measure-Object -sum).sum  # bytes
	
		# For debugging
		# Write-Host "bvr_counts_ro_total = $bvr_counts_ro_total"
		# Write-Host "bvr_sizes_ro_total = $bvr_sizes_ro_total"
		
	}
	# ... common to all folders
	$jpg_sizes_all_total = ($jpg_sizes_ary | Measure-Object -sum).sum  # bytes
	$jpg_sizes_ro_total  = ($jpg_sizes_ro_ary | Measure-Object -sum).sum  # bytes
	$mp4_sizes_all_total = ($mp4_sizes_ary | Measure-Object -sum).sum  # bytes
	$mp4_sizes_ro_total  = ($mp4_sizes_ro_ary | Measure-Object -sum).sum  # bytes
	
	# For debugging
	# Write-Host "jpg_counts_all_total = $jpg_counts_all_total"
	# Write-Host "jpg_counts_ro_total = $jpg_counts_ro_total"
	# Write-Host "mp4_counts_ro_total = $mp4_counts_ro_total"
	# Write-Host "jpg_sizes_ro_total = $jpg_sizes_ro_total"
	# Write-Host "mp4_sizes_ro_total = $mp4_sizes_ro_total"

	if ($i = 0) {
		if ($fromConsole) {
			Write-Host "Note: the tables below are sorted on the 'Total' file size column (descending)"
		}
	}

	if ($fromConsole) {
		Write-Host ""
		Write-Host "/////////////////////////////////////////////////////////////////////////////////"
		Write-Host ("Folder: " + $source)
		if ($DevMode) {
			Write-Host ""
			Write-Host ("(dev mode) folder read time = " + ($get_time_ms.elapsed.totalmilliseconds).ToString('0.00') + " msec") -ForegroundColor Cyan
		}
	}
	
	# Get last folder name in path $source - used in output.csv filename
	$folder = $source | split-path -leaf
	
	# For debugging
	#Write-Host "folder = $folder"
	
	$output_path = $app_path

	# DISPLAY TABLES.
	if ($files_ary.count -eq 0) {
		Write-Host "No Files"		
	} else {		
		# Handle the Alerts folder
		if ($source -eq $alerts_folder_path) {
			
			if ($jpg_sizes_ary.Count -gt 0) {				
				# ----------------------------------------------------------------------
				# JPG files
				#Write-Host "JPG FILES"
				
				$table_type = "jpg"
				$jpg_files_fcams   = @($jpg_files_tbl.GetEnumerator()   | % { "$($_.Name)" })   # array of strings
				$jpg_files_lists   = @($jpg_files_tbl.GetEnumerator()   | % { "$($_.Value)" })  # array of lists
				$jpg_sizes_lists   = @($jpg_sizes_tbl.GetEnumerator()   | % { "$($_.Value)" })  # array of lists
				$jpg_tstamps_lists = @($jpg_tstamps_tbl.GetEnumerator() | % { "$($_.Value)" })  # array of lists
			
				# Calculate #days for each file type
				$cam_ary = $jpg_files_fcams
				$days_ary = @()
				foreach ($x in $cam_ary) {
					if ($jpg_dates_tbl[$x]) {
						$tmp_ary = [array]$jpg_dates_tbl[$x]
						$tmp_ary = [array]($tmp_ary | select -Unique)  # Note: returns $null of array is empty
						if ($tmp_ary) {
							$day_cnt = $tmp_ary.Count
						} else {
							$day_cnt = 0
						}
					} else {
						$day_cnt = 0
					}
					$days_ary += $day_cnt
				}
				#Write-Host "cam_ary = $cam_ary"
				#Write-Host "days_ary = $days_ary"
			
				if ($DevMode) {
					if ($fromConsole) {
						Write-Host ""
						Write-Host "(dev mode) Table timers ($folder-$table_type)" -ForegroundColor Cyan
						Write-Host "(dev mode) ------------------------------" -ForegroundColor Cyan
					}
				}
				
				# Note: Function show_table() outputs DevMode timer results directly to the console (rather than return them for use here).
				# Why?  Attempting to return values from a function that uses a Format cmdlet (like Format-Table) is messy.
				# REF: https://stackoverflow.com/questions/36358047/how-can-i-store-output-from-format-table-for-later-use
				
				show_table $table_type $folder $SortColNo $Ascending $jpg_files_fcams $jpg_files_lists $jpg_sizes_lists $jpg_tstamps_lists $days_ary $jpg_sizes_ary  $jpg_tstamps_ary $output_path $timestamp_output_csv_files_flg $fromConsole $DevMode
			}

			if ($dat_sizes_ary.Count -gt 0) {
				# ----------------------------------------------------------------------
				# DAT files
				#Write-Host "DAT FILES"
				
				$table_type = "dat"
				$dat_files_fcams   = @($dat_files_tbl.GetEnumerator()   | % { "$($_.Name)" })   # array of strings
				$dat_files_lists   = @($dat_files_tbl.GetEnumerator()   | % { "$($_.Value)" })  # array of lists
				$dat_sizes_lists   = @($dat_sizes_tbl.GetEnumerator()   | % { "$($_.Value)" })  # array of lists
				$dat_tstamps_lists = @($dat_tstamps_tbl.GetEnumerator() | % { "$($_.Value)" })  # array of lists
			
				# Calculate #days for each file type
				$cam_ary = $dat_files_fcams
				$days_ary = @()
				foreach ($x in $cam_ary) {
					if ($dat_dates_tbl[$x]) {
						$tmp_ary = [array]$dat_dates_tbl[$x]
						$tmp_ary = [array]($tmp_ary | select -Unique)  # Note: returns $null of array is empty
						if ($tmp_ary) {
							$day_cnt = $tmp_ary.Count
						} else {
							$day_cnt = 0
						}
					} else {
						$day_cnt = 0
					}
					#Write-Host "$x day cnt = $day_cnt"
					$days_ary += $day_cnt
				}
				#Write-Host "cam_ary = $cam_ary"
				#Write-Host "days_ary = $days_ary"
			
				if ($DevMode) {
					if ($fromConsole) {
						Write-Host ""
						Write-Host "(dev mode) Table timers ($folder-$table_type)" -ForegroundColor Cyan
						Write-Host "(dev mode) ------------------------------" -ForegroundColor Cyan
					}
				}

				show_table $table_type $folder $SortColNo $Ascending $dat_files_fcams $dat_files_lists $dat_sizes_lists $dat_tstamps_lists $days_ary $dat_sizes_ary $dat_tstamps_ary $output_path $timestamp_output_csv_files_flg $fromConsole $DevMode
			}
			
			if ($mp4_sizes_ary.Count -gt 0) {
				# ----------------------------------------------------------------------
				# MP4 files
				#Write-Host "MP4 FILES"
				
				$table_type = "mp4"
				$mp4_files_fcams   = @($mp4_files_tbl.GetEnumerator()   | % { "$($_.Name)" })   # array of strings
				$mp4_files_lists   = @($mp4_files_tbl.GetEnumerator()   | % { "$($_.Value)" })  # array of lists
				$mp4_sizes_lists   = @($mp4_sizes_tbl.GetEnumerator()   | % { "$($_.Value)" })  # array of lists
				$mp4_tstamps_lists = @($mp4_tstamps_tbl.GetEnumerator() | % { "$($_.Value)" })  # array of lists
			
				# Calculate #days for each file type
				$cam_ary = $mp4_files_fcams
				$days_ary = @()
				foreach ($x in $cam_ary) {
					if ($mp4_dates_tbl[$x]) {
						$tmp_ary = [array]$mp4_dates_tbl[$x]
						$tmp_ary = [array]($tmp_ary | select -Unique)  # Note: returns $null of array is empty
						if ($tmp_ary) {
							$day_cnt = $tmp_ary.Count
						} else {
							$day_cnt = 0
						}
					} else {
						$day_cnt = 0
					}
					$days_ary += $day_cnt
				}
				#Write-Host "cam_ary = $cam_ary"
				#Write-Host "days_ary = $days_ary"
			
				if ($DevMode) {
					if ($fromConsole) {
						Write-Host ""
						Write-Host "(dev mode) Table timers ($folder-$table_type)" -ForegroundColor Cyan
						Write-Host "(dev mode) ------------------------------" -ForegroundColor Cyan
					}
				}

				show_table $table_type $folder $SortColNo $Ascending $mp4_files_fcams $mp4_files_lists $mp4_sizes_lists $mp4_tstamps_lists $days_ary $mp4_sizes_ary $mp4_tstamps_ary $output_path $timestamp_output_csv_files_flg $fromConsole $DevMode
			}
			
			# Show READ-ONLY file stats
			$cnt_all_total  = $dat_counts_all_total + $jpg_counts_all_total + $mp4_counts_all_total;
			$cnt_ro_total   = $dat_counts_ro_total + $jpg_counts_ro_total + $mp4_counts_ro_total;
			$size_all_total = $dat_sizes_all_total + $jpg_sizes_all_total + $mp4_sizes_all_total;
			$size_ro_total  = $dat_sizes_ro_total + $jpg_sizes_ro_total + $mp4_sizes_ro_total;
			Write-Host ""
			Write-Host ("Folder All files         = " + $cnt_all_total + " files, {0:N2} GB" -f ($size_all_total / 1GB))
			Write-Host ("Folder Read-Only files   = " + $cnt_ro_total + " files, {0:N2} GB" -f ($size_ro_total / 1GB))
			if ($size_all_total -gt 0) {
				$size_ratio = $size_ro_total/$size_all_total;
				Write-Host ("Percent Read-Only (size) = {0:P2}" -f $size_ratio)
			}

		}
		
		# Handle the non-Alerts folders
		if ($source -ne $alerts_folder_path) {
			
			if ($bvr_sizes_ary.Count -gt 0) {				
				# ----------------------------------------------------------------------
				# BVR files
				#Write-Host "BVR FILES"
				
				$table_type = "bvr"
				$bvr_files_fcams   = @($bvr_files_tbl.GetEnumerator()   | % { "$($_.Name)" })   # array of strings
				$bvr_files_lists   = @($bvr_files_tbl.GetEnumerator()   | % { "$($_.Value)" })  # array of lists
				$bvr_sizes_lists   = @($bvr_sizes_tbl.GetEnumerator()   | % { "$($_.Value)" })  # array of lists
				$bvr_tstamps_lists = @($bvr_tstamps_tbl.GetEnumerator() | % { "$($_.Value)" })  # array of lists

				# Write-Host ("bvr_files_tbl ary = " + $bvr_files_ary)
				# Write-Host ("bvr_files_fcams = " + $bvr_files_fcams)
				# Write-Host ("bvr_files_fcams cnt = " + $bvr_files_fcams.Length)
				# Write-Host ("bvr_files_lists cnt = " + $bvr_files_lists.Length)
				# Write-Host ("bvr_files_lists type = " + $bvr_files_lists.GetType())
				# Write-Host ("bvr_sizes_lists cnt = " + $bvr_sizes_lists.Length)
				# Write-Host ("bvr_sizes_lists type = " + $bvr_sizes_lists.GetType())
				# Write-Host ("bvr_sizes_vals = " + $bvr_sizes_vals)
				# Write-Host ("bvr_tstamps_vals = " + $bvr_tstamps_vals)
			
				# Calculate #days for each file type
				$cam_ary = $bvr_files_fcams
				$days_ary = @()
				foreach ($x in $cam_ary) {
					if ($bvr_dates_tbl[$x]) {
						$tmp_ary = [array]$bvr_dates_tbl[$x]
						$tmp_ary = [array]($tmp_ary | select -Unique)  # Note: returns $null of array is empty
						if ($tmp_ary) {
							$day_cnt = $tmp_ary.Count
						} else {
							$day_cnt = 0
						}
					} else {
						$day_cnt = 0
					}
					$days_ary += $day_cnt
				}
				#Write-Host "cam_ary = $cam_ary"
				#Write-Host "days_ary = $days_ary"
			
				if ($DevMode) {
					if ($fromConsole) {
						Write-Host ""
						Write-Host "(dev mode) Table timers ($folder-$table_type)" -ForegroundColor Cyan
						Write-Host "(dev mode) ------------------------------" -ForegroundColor Cyan
					}
				}

				show_table $table_type $folder $SortColNo $Ascending $bvr_files_fcams $bvr_files_lists $bvr_sizes_lists $bvr_tstamps_lists $days_ary $bvr_sizes_ary $bvr_tstamps_ary $output_path $timestamp_output_csv_files_flg $fromConsole $DevMode
			}

			if ($mp4_sizes_ary.Count -gt 0) {				
				# ----------------------------------------------------------------------
				# MP4 files
				#Write-Host "MP4 FILES"
				
				$table_type = "mp4"
				$mp4_files_fcams   = @($mp4_files_tbl.GetEnumerator()   | % { "$($_.Name)" })   # array of strings
				$mp4_files_lists   = @($mp4_files_tbl.GetEnumerator()   | % { "$($_.Value)" })  # array of lists
				$mp4_sizes_lists   = @($mp4_sizes_tbl.GetEnumerator()   | % { "$($_.Value)" })  # array of lists
				$mp4_tstamps_lists = @($mp4_tstamps_tbl.GetEnumerator() | % { "$($_.Value)" })  # array of lists
			
				# Calculate #days for each file type
				$cam_ary = $mp4_files_fcams
				$days_ary = @()
				foreach ($x in $cam_ary) {
					if ($mp4_dates_tbl[$x]) {
						$tmp_ary = [array]$mp4_dates_tbl[$x]
						$tmp_ary = [array]($tmp_ary | select -Unique)  # Note: returns $null of array is empty
						if ($tmp_ary) {
							$day_cnt = $tmp_ary.Count
						} else {
							$day_cnt = 0
						}
					} else {
						$day_cnt = 0
					}
					$days_ary += $day_cnt
				}
				#Write-Host "cam_ary = $cam_ary"
				#Write-Host "days_ary = $days_ary"
			
				if ($DevMode) {
					if ($fromConsole) {
						Write-Host ""
						Write-Host "(dev mode) Table timers ($folder-$table_type)" -ForegroundColor Cyan
						Write-Host "(dev mode) ------------------------------" -ForegroundColor Cyan
					}
				}

				show_table $table_type $folder $SortColNo $Ascending $mp4_files_fcams $mp4_files_lists $mp4_sizes_lists $mp4_tstamps_lists $days_ary $mp4_sizes_ary $mp4_tstamps_ary $output_path $timestamp_output_csv_files_flg $fromConsole $DevMode
			}

			if ($jpg_sizes_ary.Count -gt 0) {				
				# ----------------------------------------------------------------------
				# JPG files
				#Write-Host "JPG FILES"
				
				$table_type = "jpg"
				$jpg_files_fcams   = @($jpg_files_tbl.GetEnumerator()   | % { "$($_.Name)" })   # array of strings
				$jpg_files_lists   = @($jpg_files_tbl.GetEnumerator()   | % { "$($_.Value)" })  # array of lists
				$jpg_sizes_lists   = @($jpg_sizes_tbl.GetEnumerator()   | % { "$($_.Value)" })  # array of lists
				$jpg_tstamps_lists = @($jpg_tstamps_tbl.GetEnumerator() | % { "$($_.Value)" })  # array of lists
			
				# Calculate #days for each file type
				$cam_ary = $jpg_files_fcams
				$days_ary = @()
				foreach ($x in $cam_ary) {
					if ($jpg_dates_tbl[$x]) {
						$tmp_ary = [array]$jpg_dates_tbl[$x]
						$tmp_ary = [array]($tmp_ary | select -Unique)  # Note: returns $null of array is empty
						if ($tmp_ary) {
							$day_cnt = $tmp_ary.Count
						} else {
							$day_cnt = 0
						}
					} else {
						$day_cnt = 0
					}
					#Write-Host "$x day cnt = $day_cnt"
					$days_ary += $day_cnt
				}
				#Write-Host "cam_ary = $cam_ary"
				#Write-Host "days_ary = $days_ary"
			
				if ($DevMode) {
					if ($fromConsole) {
						Write-Host ""
						Write-Host "(dev mode) Table timers ($folder-$table_type)" -ForegroundColor Cyan
						Write-Host "(dev mode) ------------------------------" -ForegroundColor Cyan
					}
				}

				show_table $table_type $folder $SortColNo $Ascending $jpg_files_fcams $jpg_files_lists $jpg_sizes_lists $jpg_tstamps_lists $days_ary $jpg_sizes_ary $jpg_tstamps_ary $output_path $timestamp_output_csv_files_flg $fromConsole $DevMode
			}
			
			# Show READ-ONLY file stats
			$cnt_all_total  = $bvr_counts_all_total + $jpg_counts_all_total + $mp4_counts_all_total;
			$cnt_ro_total   = $bvr_counts_ro_total + $jpg_counts_ro_total + $mp4_counts_ro_total;
			$size_all_total = $bvr_sizes_all_total + $mp4_sizes_all_total + $jpg_sizes_all_total;
			$size_ro_total  = $bvr_sizes_ro_total + $mp4_sizes_ro_total + $jpg_sizes_ro_total;
			Write-Host ""
			Write-Host ("Folder All files         = " + $cnt_all_total + " files, {0:N2} GB" -f ($size_all_total / 1GB))
			Write-Host ("Folder Read-Only files   = " + $cnt_ro_total + " files, {0:N2} GB" -f ($size_ro_total / 1GB))
			if ($size_all_total -gt 0) {
				$size_ratio = $size_ro_total/$size_all_total;
				Write-Host ("Percent Read-Only (size) = {0:P2}" -f $size_ratio)
			}

		}

		$i++
	}	
}

# PERFORM MAINTENANCE ON FILE ARCHIVE FOLDERS
if ($fromConsole) {
	Write-Host ""
	Write-Host "/////////////////////////////////////////////////////////////////////////////////"
	Write-Host "Archived CSV Files Maintenance Summary:"
	#Write-Host "---------------------------------------"
}

if ($archive_lifetime_days -ne 0) {
	
	$purge_filespec = $output_path + $PathSep + "*_20*.csv"

	# Returns... @(found_files_cnt, purged_files_cnt, oldest_file_path, oldest_file_lastwritetime, oldest_file_age_days)
	#            or (string) 'ERR/' message
	$arr = `
		PurgeOldFiles `
		-Pattern $purge_filespec `
		-Days $archive_lifetime_days
		
	# Write-Host "purge_filespec = '$purge_filespec'"
	# Write-Host ("arr type = " + $arr.GetType())
	# Write-Host ("arr cnt = " + $arr.Count)
	# Write-Host ("arr = " + ($arr | ConvertTo-Json))

	if (!($arr.GetType() -match 'string')) {
		if ($fromConsole) {
			Write-Host ("Count of archived CSV files matching '$purge_filespec' = " + $arr[0])
			Write-Host ("- Oldest archived file: '" + $arr[2] + "'")
			Write-Host ("- Oldest archived file date modified: " + $arr[3])
			Write-Host ("- Oldest archived file age = " + $arr[4] + " days (before pruning)")
			#if ($timestamp_output_csv_files_flg -and $archive_lifetime_days -gt 0) {
			if ($timestamp_output_csv_files_flg -and $archive_lifetime_days -gt 0) {  # suggested by M Hotchin 5/16/2024
				Write-Host ("Count of archived CSV files pruned (age >= $archive_lifetime_days days) = " + $arr[1])
			}
			if ($archive_lifetime_days -lt 0) {
				Write-Host "All timestamped CSV files deleted per argument -ClearArchive"
			}
		}
	} else {
		if ($fromConsole) {
			Write-Host $arr
		} else {
			$rtn_status_str = 'Warning - CSV file archive managment issue'
		}
	}
	
} else {
	if ($fromConsole) {
		Write-Host "Maintenance disabled by argument -NoPruning"
	}		
}

$timer_script.stop()

if ($fromConsole) {
	if ($DevMode) {
		Write-Host ""
		Write-Host ("(dev mode) script run time = " + ($timer_script.elapsed.totalseconds).ToString("0.00") + " sec") -ForegroundColor Cyan
	}
	Write-Host ""
} else {
	return $rtn_status_str
}

#Stop-Transcript